/*
 * Physical mapping layer for MTD using the Axis partitiontable format
 *
 * Copyright (c) 2001-2006 Axis Communications AB
 *
 * This file is under the GPL.
 *
 * First partition is always sector 0 regardless of if we find a partitiontable
 * or not. In the start of the next sector, there can be a partitiontable that
 * tells us what other partitions to define. If there isn't, we use a default
 * partition split defined below.
 *
 * $Log: axisflashmap.c,v $
 * Revision 1.10  2006/12/11 15:51:05  ricardw
 * Changed name of CONFIG_ETRAX_PTABLE_SECTOR and moved it to Kconfig instead of
 * local define in axisflashmap.c .
 * Added CONFIG_ARGUS_MTD0WHOLE to set up mtd0 as whole flash.
 *
 * Revision 1.9  2006/11/28 17:24:18  bjornw
 * Dont hardcode GPIO_nCS5 for NAND NCE, because CS5 does only exist on Argus-3. Fix this better later..
 *
 * Revision 1.8  2006/11/03 10:13:19  pkj
 * Avoid compiler warnings.
 *
 * Revision 1.7  2006/07/21 12:18:15  bjornw
 * Added hardware ECC support, default partitions and accelerated block-read/write for NAND-flashes
 *
 * Revision 1.6  2006/05/15 14:53:12  bjornw
 * Added a comment
 *
 * Revision 1.5  2005/09/02 17:30:39  bjornw
 * Experimental NAND-flash support
 *
 * Revision 1.4  2005/07/29 17:44:04  bjornw
 * Dont crash if there is no romfs at all
 *
 * Revision 1.3  2005/06/13 19:55:38  bjornw
 * Do cfi_probe_init before probing cfi. I dont know why this is still necessary but it is.
 *
 * Revision 1.2  2005/06/07 16:14:30  bjornw
 * 2.6 version
 *
 * Revision 1.1  2005/06/07 13:33:19  bjornw
 * Argus for linux 2.6 initial temptations
 *
 *
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>

#include <linux/mtd/concat.h>
#include <linux/mtd/map.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/partitions.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/nand_ecc.h>
#include <linux/mtd/mtdram.h>

#include <asm/io.h>
#include <asm/arch/axisflashmap.h>
#include <asm/arch/argus.h>
#include <asm/arch/gpio.h>
#include <asm/arch/argus_addr.h>

// Debug NAND-flash (low-level!)
#define DN(x)

#define FLASH_UNCACHED_ADDR         EXT0_NOCACHE_V
#define FLASH_CACHED_ADDR           EXT0_CACHE_V

// CMD and DATA are decoded through A18 and A19 respectively, within the
// same 1 MB mapping as NAND_NOCACHE_DATA_V.
#define NAND_NOCACHE_CMD_V    (NAND_NOCACHE_DATA_V + 0x40000)
#define NAND_NOCACHE_ADDR_V   (NAND_NOCACHE_DATA_V + 0x80000)

/*
 * WINDOW_SIZE is the total size where the flash chips may be mapped.
 * MTD probes should find all devices there and it does not matter
 * if there are unmapped gaps or aliases (mirrors of flash devices).
 * The MTD probes will ignore them.
 */

#define WINDOW_SIZE  (32 * 1024 * 1024)

int cfi_probe_init(void);

/* Initialize, because these are set BEFORE the BSS is cleared and
 * so should not end up in BSS themselves...
 */

unsigned long romfs_start = 42, romfs_length = 42,
	romfs_in_flash = 42; /* set in head-armv.S */

/* The master mtd for the entire flash. */
struct mtd_info* axisflash_mtd = NULL;

#if CONFIG_ARGUS_FLASH_BUSWIDTH==1
#define flash_data __u8
#elif CONFIG_ARGUS_FLASH_BUSWIDTH==2
#define flash_data __u16
#elif CONFIG_ARGUS_FLASH_BUSWIDTH==4
#define flash_data __u32
#endif

/* 
 * NOR-flash map driver
 *
 * We run into tricky coherence situations if we mix cached with uncached
 * accesses to we use the uncached version here.
 */

static map_word flash_read(struct map_info *map, unsigned long ofs)
{
	map_word tmp;
	tmp.x[0] = *(flash_data *)(FLASH_UNCACHED_ADDR + ofs);
	return tmp;
}

static void flash_copy_from(struct map_info *map, void *to,
			    unsigned long from, ssize_t len)
{
	memcpy(to, (void *)(FLASH_UNCACHED_ADDR + from), len);
}

static void flash_write(struct map_info *map, map_word d, unsigned long adr)
{
	*(flash_data *)(FLASH_UNCACHED_ADDR + adr) = (flash_data)d.x[0];
}

/* NAND-flash basic functions
 *
 */

#ifdef CONFIG_MTD_NAND

static int 
argus_device_ready(struct mtd_info *mtd)
{
	// NAND RDY/_BSY is connected to GPIO_LED on Axis 218/219 (at least)
	// (make this selectable). DREQ0 on Olympus-3!

	return gpio_get_level(GPIO_DREQ0);
}

#if defined(CONFIG_ARGUS_3)
#define GPIO_NCE GPIO_nCS5
#else
// TODO! What pin is used on Argus-4 ?
#define GPIO_NCE 0
#endif

static void 
argus_hwcontrol(struct mtd_info *mtd, int cmd)
{
        struct nand_chip *chip = mtd->priv;

        switch (cmd) {
		// Manually control the chip-select pin. This is possible because
		// Argus gates the NAND_RE/NAND_WE outputs internally with the NCE pin before
		// it goes to the GPIO-mux (and NCE is normally a chip-select pin).
		case NAND_CTL_SETNCE:
			DN(printk("SET NCE\n"));
			gpio_set_mode(GPIO_NCE, GPIO_MODE_OUT, 0);
			break;
		case NAND_CTL_CLRNCE:
			DN(printk("CLR NCE\n"));
			gpio_set_mode(GPIO_NCE, GPIO_MODE_OUT, 1);
			break;
			
		// Choose ADDR, CMD or DATA addressing spaces
		case NAND_CTL_SETCLE:
			DN(printk("SET CLE\n"));
			chip->IO_ADDR_W = chip->IO_ADDR_R = NAND_NOCACHE_CMD_V;
			break;
			
		case NAND_CTL_SETALE:
			DN(printk("SET ALE\n"));
			chip->IO_ADDR_W = chip->IO_ADDR_R = NAND_NOCACHE_ADDR_V;
			break;
			
			/* NAND_CTL_CLRCLE: */
			/* NAND_CTL_CLRALE: */
		default:
			DN(printk("CLRCLE/CLRALE\n"));
			chip->IO_ADDR_W = chip->IO_ADDR_R = NAND_NOCACHE_DATA_V;
			break;
        }
}

/* override the standard functions for a little more speed. We can
 * use read/write block to move the data buffers to/from the controller
 */

static void 
argus_read_buf(struct mtd_info *mtd, unsigned char *buf, int len)
{
        struct nand_chip *this = mtd->priv;
        readsb(this->IO_ADDR_R, buf, len);
}

static void
argus_write_buf(struct mtd_info *mtd,
		const unsigned char *buf, int len)
{
        struct nand_chip *this = mtd->priv;
        writesb(this->IO_ADDR_W, buf, len);
}

#define ARGUS3_ECC_COMPAT(ecc) (((ecc)^~0x30000) & 0xffffff)
// Compared to the Argus III HW ECC calculation, the Linux SW 
// variant has swapped the least significant and the "middle 
// significant" bytes and the ECC from the HW need to be converted
// using the define above to be exact the same as the Linux SW ECC

static void 
argus_enable_hwecc(struct mtd_info* mtd, int mode)
{
	//printk("Argus HWECC %d\n", mode);
        SOC_EXTMEM.NANDCLEAR = 1;
}

static int 
argus_calculate_ecc(struct mtd_info* mtd, const unsigned char* dat,
		    unsigned char* ecc_code)
{
        unsigned int ecc = ARGUS3_ECC_COMPAT(SOC_EXTMEM.NANDECC0);
        ecc_code[1] = ecc;
        ecc_code[0] = ecc >> 8; 
        ecc_code[2] = ecc >> 16;
//      printk("Argus ECC 0x%x (0x%x)\n", ARGUS3_ECC_COMPAT(SOC_EXTMEM.NANDECC0), SOC_EXTMEM.NANDECC0);
//      printk("            %x %x %x \n", ecc_code[0], ecc_code[1], ecc_code[2]);
        return 1;
}

#define NUM_DEFAULT_NAND_PARTITIONS 3
static struct mtd_partition axis_default_nand_partitions[NUM_DEFAULT_NAND_PARTITIONS] = {
	{
		name: "boot",
		size: 16*1024,  // 16 kb Argus bootblock
		offset: 0
	},
	{
		name: "root",
		size: 0x1000000,  // 16 Mb
		offset: MTDPART_OFS_NXTBLK
	},
	{
		name: "user",
		size: 0x3000000 - (16 * 1024),  // 48 Mb
		offset: MTDPART_OFS_NXTBLK
	}
};

#endif

#ifdef CONFIG_ARGUS_AXISFLASHMAP_MTD0WHOLE
/* Main flash device */
static struct mtd_partition main_partition = {
	.name = "main",
	.size = 0,
	.offset = 0
};
#endif

// TODO: set .virt and .cached to let MTD handle the accesses itself, we don't
// really need to go through any advanced mappings here. Do it like this for
// now though.

static struct map_info axis_map = {
	name: "Axis flash",
	size: WINDOW_SIZE,
	bankwidth: CONFIG_ARGUS_FLASH_BUSWIDTH,
	read: flash_read,
	copy_from: flash_copy_from,
	write: flash_write
};

/* If no partition-table was found, we use this default-set.
 */

#define MAX_PARTITIONS         7  
#define NUM_DEFAULT_PARTITIONS 3

/* Default flash size is 2MB. CONFIG_ARGUS_PTABLE_SECTOR is most likely the
 * size of one flash block and "filesystem"-partition needs 5 blocks to be able
 * to use JFFS.
 */
static struct mtd_partition axis_default_partitions[NUM_DEFAULT_PARTITIONS] = {
	{
		name: "boot firmware",
		size: CONFIG_ARGUS_PTABLE_SECTOR,
		offset: 0
	},
	{
		name: "kernel",
		size: 0x200000 - (6 * CONFIG_ARGUS_PTABLE_SECTOR),
		offset: CONFIG_ARGUS_PTABLE_SECTOR
	},
	{
		name: "filesystem",
		size: 5 * CONFIG_ARGUS_PTABLE_SECTOR,
		offset: 0x200000 - (5 * CONFIG_ARGUS_PTABLE_SECTOR)
	}
};

static struct mtd_partition axis_partitions[MAX_PARTITIONS] = {
	{
		name: "part0",
		size: CONFIG_ARGUS_PTABLE_SECTOR,
		offset: 0
	},
	{
		name: "part1",
		size: 0,
		offset: 0
	},
	{
		name: "part2",
		size: 0,
		offset: 0
	},
	{
		name: "part3",
		size: 0,
		offset: 0
	},
	{
		name: "part4",
		size: 0,
		offset: 0
	},
	{
		name: "part5",
		size: 0,
		offset: 0
	},
	{
		name: "part6",
		size: 0,
		offset: 0
	},
};

/* CFI-scan the flash, and if there was a chip, read the partition-table
 * and register the partitions with MTD.
 */

static int __init
init_axis_flash(void)
{
	struct mtd_info *mymtd = NULL;
	int err = 0;
	int pidx = 0;
	struct partitiontable_head *ptable_head = NULL;
	struct partitiontable_entry *ptable;
	int use_default_ptable = 1; /* Until proven otherwise */
	const char *pmsg = "  /dev/flash%d at 0x%x, size 0x%x\n";

	printk(KERN_NOTICE "Axis flash mapping: %x at %x\n",
	       WINDOW_SIZE, FLASH_CACHED_ADDR);

#ifdef CONFIG_MTD_CFI
	if (!mymtd) {
		// The cfi module is linked after this in the linker
		// so its init function has not been called yet.
		cfi_probe_init();
		printk("trying to cfi_probe...\n");
		mymtd = (struct mtd_info *)do_map_probe("cfi_probe", &axis_map);
	}
#endif

	if(mymtd) {
		printk(KERN_INFO "%s: 0x%08x bytes of flash memory.\n",
		       mymtd->name, mymtd->size);
		axisflash_mtd = mymtd;
		mymtd->owner = THIS_MODULE;
		ptable_head = (struct partitiontable_head *)
			(FLASH_CACHED_ADDR + CONFIG_ARGUS_PTABLE_SECTOR + 
			 PARTITION_TABLE_OFFSET);
	}
	pidx++;  /* first partition is always set to the default */

	if (ptable_head && (ptable_head->magic == PARTITION_TABLE_MAGIC)
	    && (ptable_head->size <
		(MAX_PARTITIONS * sizeof(struct partitiontable_entry) +
		PARTITIONTABLE_END_MARKER_SIZE))
	    && (*(unsigned long*)((void*)ptable_head + sizeof(*ptable_head) +
				  ptable_head->size -
				  PARTITIONTABLE_END_MARKER_SIZE)
		== PARTITIONTABLE_END_MARKER)) {
		/* Looks like a start, sane length and end of a
		 * partition table, lets check csum etc.
		 */
		int ptable_ok = 0;
		struct partitiontable_entry *max_addr =
			(struct partitiontable_entry *)
			((unsigned long)ptable_head + sizeof(*ptable_head) +
			 ptable_head->size);
		unsigned long offset = CONFIG_ARGUS_PTABLE_SECTOR;
		unsigned char *p;
		unsigned long csum = 0;
                
		ptable = (struct partitiontable_entry *)
			((unsigned long)ptable_head + sizeof(*ptable_head));

		/* Lets be PARANOID, and check the checksum. */
		p = (unsigned char*) ptable;

		while (p <= (unsigned char*)max_addr) {
			csum += *p++;
			csum += *p++;
			csum += *p++;
			csum += *p++;
		}
		/* printk("  total csum: 0x%08X 0x%08X\n",
		   csum, ptable_head->checksum); */
		ptable_ok = (csum == ptable_head->checksum);

		/* Read the entries and use/show the info.  */
		printk(" Found %s partition table at 0x%08lX-0x%08lX.\n",
		       (ptable_ok ? "valid" : "invalid"),
		       (unsigned long)ptable_head,
		       (unsigned long)max_addr);

		/* We have found a working bootblock.  Now read the
		   partition table.  Scan the table.  It ends when
		   there is 0xffffffff, that is, empty flash.  */
		
		while (ptable_ok
		       && ptable->offset != 0xffffffff
		       && ptable < max_addr
		       && pidx < MAX_PARTITIONS) {

			axis_partitions[pidx].offset = offset + ptable->offset;
			axis_partitions[pidx].size = ptable->size;

			printk(pmsg, pidx, axis_partitions[pidx].offset,
			       axis_partitions[pidx].size);
			pidx++;
			ptable++;
		}
		use_default_ptable = !ptable_ok;
	}

	if (romfs_in_flash) {
		/* Add an overlapping device for the root partition (romfs). */

		axis_partitions[pidx].name = "romfs";
		axis_partitions[pidx].size = romfs_length;
		axis_partitions[pidx].offset = romfs_start - 0x80000000;
		axis_partitions[pidx].mask_flags |= MTD_WRITEABLE;

		printk(KERN_INFO 
		       " Adding readonly flash partition for romfs image:\n");
		printk(pmsg, pidx, axis_partitions[pidx].offset,
		       axis_partitions[pidx].size);
		pidx++;
	}

#ifdef CONFIG_ARGUS_AXISFLASHMAP_MTD0WHOLE
	if (mymtd) {
		main_partition.size = mymtd->size;
                err = add_mtd_partitions(mymtd, &main_partition, 1);
		if (err)
			panic("axisflashmap: Could not initialize "
			      "partition for whole main mtd device!\n");
	}
#endif

	if(mymtd) {
		if (use_default_ptable) {
			printk(KERN_INFO " Using default partition table.\n");
			err = add_mtd_partitions(mymtd, axis_default_partitions,
						 NUM_DEFAULT_PARTITIONS);
		} else {
			err = add_mtd_partitions(mymtd, axis_partitions, pidx);
		}
		
		if (err) {
			panic("axisflashmap could not add MTD partitions!\n");
		}
	}

	if (!romfs_in_flash && romfs_start != 42) {
		/* Create an RAM device for the root partition (romfs). */

#if !defined(CONFIG_MTD_MTDRAM) || (CONFIG_MTDRAM_TOTAL_SIZE != 0) || (CONFIG_MTDRAM_ABS_POS != 0)
		/* No use trying to boot this kernel from RAM. Panic! */
		printk(KERN_EMERG "axisflashmap: Cannot create an MTD RAM "
		       "device due to kernel (mis)configuration!\n");
		panic("This kernel cannot boot from RAM!\n");
#else
		struct mtd_info *mtd_ram;

		mtd_ram = (struct mtd_info *)kmalloc(sizeof(struct mtd_info),
						     GFP_KERNEL);
		if (!mtd_ram) {
			panic("mtd_info oom\n");
		}

		printk(KERN_INFO " Adding RAM partition for romfs image (magic 0x%lx):\n",
		       *(unsigned long*)(romfs_start));
		printk(pmsg, pidx, (unsigned)romfs_start, (unsigned)romfs_length);
		// This does add_mtd_device
		err = mtdram_init_device(mtd_ram, (void*)romfs_start, 
		                         romfs_length, "romfs");
		if (err) {
			panic("mtd_ram init failed\n");
		}
#endif
	}

#if defined(CONFIG_MTD_NAND)
	{
		struct mtd_info *mtd_nand;
		struct nand_chip *chip;
		int err;

		mtd_nand = (struct mtd_info *)kmalloc(sizeof(struct mtd_info) + 
						      sizeof(struct nand_chip),
						     GFP_KERNEL);
		if (!mtd_nand) {
			panic("mtd_nand oom!\n");
		}
		// Clear everything
		memset((char *) mtd_nand, 0, sizeof(struct mtd_info) + sizeof(struct nand_chip));

		SOC_EXTMEM.CONFIG[5] = SET_EXTCONFIG_WIDTH(0) | SET_EXTCONFIG_SETUP(0) | SET_EXTCONFIG_WAIT(3);

		// nand_chip struct is just after mtd_nand
		chip = (struct nand_chip *) (&mtd_nand[1]);
		mtd_nand->priv = chip;
		chip->IO_ADDR_R    = NAND_NOCACHE_DATA_V;
		chip->IO_ADDR_W    = NAND_NOCACHE_DATA_V;
		chip->chip_delay   = 25; // not used if we provide dev_ready to poll for readiness
		//chip->options      = NAND_NO_AUTOINCR;
		chip->write_buf    = argus_write_buf;
		chip->read_buf     = argus_read_buf;
		chip->dev_ready    = argus_device_ready;
		chip->hwcontrol    = argus_hwcontrol;

		chip->eccmode = NAND_ECC_HW3_256;
		chip->enable_hwecc = argus_enable_hwecc;
		chip->calculate_ecc = argus_calculate_ecc;
		chip->correct_data = nand_correct_data;  // default algorithm

		//chip->eccmode = NAND_ECC_SOFT;

		// Scan to find existence of the device
		printk("scanning for NAND device...\n");
		if (nand_scan (mtd_nand, 1)) {
			goto nonand;
		}

		printk("adding default NAND mtd device partitions...\n");
		add_mtd_partitions(mtd_nand, axis_default_nand_partitions, NUM_DEFAULT_NAND_PARTITIONS);
#if 0
		err = add_mtd_device(mtd_nand);
		if (err) {
			panic("add_mtd_dev of nand dev failed\n");
		}
#endif
	}
 nonand:
#endif


	return err;
}

/* This adds the above to the kernels init-call chain */

module_init(init_axis_flash);

EXPORT_SYMBOL(axisflash_mtd);
