/*
 * CRYPTO crypto driver for Artpec-B
 *
 * Copyright (C) 2007  Axis Communications AB
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <linux/interrupt.h>
#include <linux/mutex.h>

#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/arch/argdma.h>
#include <asm/arch/dma_alloc.h>
#include <asm/arch/crypto.h>
#include <asm/arch/argus4.h>
#include <asm/arch/ciobuf.h>

/* The device number. */
#define CRYPTO_MAJOR    (254)
#define CRYPTO_MINOR    (0)

#define LDEBUG
#ifdef LDEBUG
#define DEBUG(s) s
#define DEBUG_API(s) s
#define assert(s) do{if (!(s)) panic(#s);} while(0);
#else
#define DEBUG(s)
#define assert(s)
#endif

#define CRYPTO_MAX_KEY_LENGTH      (32)

/* Next Session ID to assign. */
static crypto_session_id next_sid = 1;

struct crypto_ctrl
{
	int dma_in;
	int dma_out;
	struct mutex lock;
	struct completion done;
	struct ciobuf* ciobuf_in;
	struct ciobuf* ciobuf_out;
};

struct crypto_session{
	crypto_session_id sid;
	struct crypto_session *next;
	crypto_cipher_mode     cmode; 
	unsigned char          key[CRYPTO_MAX_KEY_LENGTH];
	unsigned int           keylen;
};

struct crypto_private{
        struct crypto_session *sessions;
};

static struct crypto_ctrl ctrl;

static int init_crypto(void);

static int crypto_open(struct inode *inode, struct file *filp)
{
	if (iminor(inode) != CRYPTO_MINOR) 
		return -EINVAL;
	filp->private_data = 
	  kmalloc(sizeof(struct crypto_private), GFP_KERNEL);
	if (!filp->private_data){
		DEBUG_API(printk("crypto_open, kmalloc crypto_private\n"));
		return -ENOMEM;
	}	

	return 0;
}

static int crypto_release(struct inode *inode, struct file *filp)
{
	return 0;
}

static void crypto_irq(int cha)
{
	printk("CRYPTO IRQ\n");
	complete(&ctrl.done);
}

static int crypto_ioctl_process(struct inode *inode, struct file *filp,
	             unsigned int cmd, unsigned long arg)
{
	struct crypto_private         *priv;
	struct crypto_session         *sess;
	struct crypto_process_op      op;
	int err;
	int i;
	int keysize;

	printk("Process\n");

	err = copy_from_user(&op, (void*)arg, sizeof(op));
	if (err) return -EFAULT;

	priv = (struct crypto_private*)filp->private_data;
	sess = priv->sessions;

	while (sess && sess->sid != op.ses_id) {
		sess = sess->next;    
	}
	if (!sess)
		return -EINVAL;

	printk("Keylen %d\n", sess->keylen);

	if (sess->keylen == 32)
		keysize = 2;
	else if (sess->keylen == 24)
		keysize = 1;
	else if (sess->keylen == 16)
		keysize = 0;
	else {
		return -EINVAL;
	}

	// Exclusive access to crypto hardware
	mutex_lock_interruptible(&ctrl.lock);

	// Setup AES hardware
	SOC_AES.CTRL = 
	  SET_DECIPHER_MODE(op.op == OP_DECRYPT ? 1 : 0) |
	  SET_CBC_ENABLE(sess->cmode == crypto_cipher_mode_cbc ? 1 : 0) |
	  SET_CRYPTO_KEYSIZE(keysize) |
	  SET_CRYPTO_MEMBANK(0)       |
	  SET_DMAIN_ENABLE(1)         |
	  SET_DMAOUT_ENABLE(1)        |
	  SET_JOB_ENABLE(1);

	// Setup key 
	SOC_AES.MEM[0][0] = ((unsigned*)op.cipher_iv)[0];
	SOC_AES.MEM[0][1] = ((unsigned*)op.cipher_iv)[1];
	SOC_AES.MEM[0][2] = ((unsigned*)op.cipher_iv)[2];
	SOC_AES.MEM[0][3] = ((unsigned*)op.cipher_iv)[3];
	for (i = 0; i < sess->keylen / 4; i++) {
	   SOC_AES.MEM[0][i+4] = ((unsigned*)sess->key)[i];
	}
	
	// Map data to dma
	printk("Map\n");
	ciobuf_map(&ctrl.ciobuf_in, (unsigned long)op.indata, op.inlen, READ);
	ciobuf_map(&ctrl.ciobuf_in, (unsigned long)op.outdata, op.outlen, WRITE);

	// Start DMAs
	printk("Start DMA\n");
	dma_start_ciobuf(ctrl.dma_in, ctrl.ciobuf_in, 0, 0, crypto_irq);
	dma_start_ciobuf(ctrl.dma_out, ctrl.ciobuf_out, 0, 0, NULL);

	printk("Wait\n");
	// Wait for completion
	wait_for_completion_interruptible(&ctrl.done);

	printk("Done\n");
	// Hardware free again
	mutex_unlock(&ctrl.lock);

	return 0;
}

static int crypto_ioctl_close_session(struct inode *inode, struct file *filp,
				      unsigned int cmd, unsigned long arg)
{
	struct crypto_private         *priv;
	struct crypto_session         *sess;
	struct crypto_session         *prev;
	struct crypto_session_op      tmp;
	int err;

	printk("Close\n");
	err = copy_from_user(&tmp, (void*)arg, sizeof(tmp));
	if (err) return -EFAULT;

	priv = (struct crypto_private*)filp->private_data;
	sess = priv->sessions;
	prev = NULL;

	while (sess && sess->sid != tmp.ses_id) {
		prev = sess;
		sess = sess->next;
	}
	if (!sess)
		return -EINVAL;

	if (prev)
		prev->next = sess->next;
	else
		priv->sessions = sess->next;
	   
	return 0;
}

static int crypto_ioctl_create_session(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	struct crypto_private         *priv;
	struct crypto_session         *sess;
	struct crypto_session_op      op;
	int err;

	printk("Create\n");
	err = copy_from_user(&op, (void*)arg, sizeof(op));
	if (err) return -EFAULT;

	priv = filp->private_data;
	sess = kmalloc(sizeof(struct crypto_session), GFP_KERNEL);
	if (!sess){
		DEBUG_API(printk("crypto_new_session, kmalloc crypto_session\n"));
		return -ENOMEM;
	}

	sess->sid = next_sid;
	next_sid++;
	/* TODO If we are really paranoid we should do duplicate check to handle sid wraparound.
	 *      OTOH 2^64 is a really large number of session. */
	if (next_sid == 0) next_sid = 1;

	// Copy config
	memcpy(sess->key, op.key, op.keylen);
	sess->keylen = op.keylen;
	sess->cmode = op.cmode;

	/* Prepend to session list. */
	sess->next = priv->sessions;
	priv->sessions = sess;
	return 0;
}

static int crypto_ioctl(struct inode *inode, struct file *file,
	             unsigned int cmd, unsigned long arg)
{ 
	int err = 0;

        if (_IOC_TYPE(cmd) != ARTPECBCRYPTO_IOCTYPE) {
		return -ENOTTY;
	}
	if (_IOC_NR(cmd) > CRYPTO_IO_MAXNR){
		return -ENOTTY;
	}

	if (_IOC_DIR(cmd) & _IOC_READ)
		err = !access_ok(VERIFY_WRITE, (void *)arg, _IOC_SIZE(cmd));
	if (!err && (_IOC_DIR(cmd) & _IOC_WRITE))
		err = !access_ok(VERIFY_READ, (void *)arg, _IOC_SIZE(cmd));
	if (err) return -EFAULT;

	switch (cmd) {
	case CRYPTO_IO_CREATE_SESSION:
		return crypto_ioctl_create_session(inode, file, cmd, arg);
	case CRYPTO_IO_CLOSE_SESSION:
		return crypto_ioctl_close_session(inode, file, cmd, arg);
	case CRYPTO_IO_PROCESS_OP:
		return crypto_ioctl_process(inode, file, cmd, arg);
	default:
		return -ENOTTY;
	}

	return 0;
}

struct file_operations crypto_fops = {
	owner: THIS_MODULE,
	open: crypto_open,
	release: crypto_release,
	ioctl: crypto_ioctl
};

static const char crypto_name[] = "CRYPTO";

static int init_crypto(void)
{
	int err;
	static int initialized = 0;
        
	if (initialized)
		return 0;

	initialized = 1;
	printk("Artpec-B CRYPTO driver, (c) 2007 Axis Communications AB\n");

	err = register_chrdev(CRYPTO_MAJOR, crypto_name, &crypto_fops);
	if (err < 0) {
		printk(KERN_ERR "CRYPTO: could not get major number\n");
		return err;
	}

	mutex_init(&ctrl.lock);
	init_completion(&ctrl.done);
	
	// Start clock to AES
	SOC_CLK.ENABLE |= CLK_AES;

	// Allocate DMAs
	ctrl.dma_out = dma_alloc(NULL, DMA_ALLOC_FIFO32);
	dma_setup(ctrl.dma_out, DMA_PORT_AES_OUT, 0, 0, DMA_SIZE_WORD, DMA_MEM2DEV);
	ctrl.dma_in = dma_alloc(NULL, DMA_ALLOC_FIFO32);
	dma_setup(ctrl.dma_out, DMA_PORT_AES_IN, 0, 0, DMA_SIZE_WORD, DMA_DEV2MEM);

	return 0;
}

module_init(init_crypto);

