// Copyright (c) 2003-2006 Axis Communications AB
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation.
//
// Argus driver for GPIO. Includes a special LEDS driver and PWM driver.

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/slab.h>
#include <linux/interrupt.h>

#if defined(CONFIG_ARGUS_MEMORY_MAPPED_GPIO)
#include <asm-arm/io.h>
#endif

#include <asm/arch/gpiodriver.h>
#include <asm/arch/gpio.h> /* To get dependency right */

#define GPIO_MAJOR 120  /* experimental MAJOR number */
#define PORT_BITS   32  /* Number of bits serverd by each minor deive */
#define LED_MINOR GPIO_MINOR_LEDS

/* disable_irq() and enable_irq() does extra unwanted checks */
#define gpio_irq_save_disable(flags) do { \
  unsigned long irqflags; \
  local_irq_save(irqflags); \
  flags = SOC_IRQ.MASK & (IRQ_GPIO_IRQ_MASK); \
  SOC_IRQ.MASK &= ~ IRQ_GPIO_IRQ_MASK; \
  local_irq_restore(irqflags); \
} while(0)

#define gpio_irq_restore(flags) do { \
  unsigned long irqflags; \
  local_irq_save(irqflags); \
  SOC_IRQ.MASK |= flags; \
  local_irq_restore(irqflags); \
} while(0)

#define gpio_irq_enable() do { \
  unsigned long irqflags; \
  local_irq_save(irqflags); \
  SOC_IRQ.MASK |= IRQ_GPIO_IRQ_MASK; \
  local_irq_restore(irqflags); \
} while(0)

#define gpio_irq_disable() do { \
  unsigned long irqflags; \
  local_irq_save(irqflags); \
  SOC_IRQ.MASK &= ~IRQ_GPIO_IRQ_MASK; \
  local_irq_restore(irqflags); \
} while(0)

#if 0
#define D(x) do { printk("%lu: ", jiffies); x; } while(0)
#define DINT(x) //do { printk("%lu: ", jiffies); x; } while(0)
#else
#define D(x)
#define DINT(x)
#endif

struct gpio_private {
       struct gpio_private* next;
       unsigned long highalarm, lowalarm;
       unsigned long changeable_bits;
       wait_queue_head_t alarm_wq;
       int base; /* First GPIO bit served by this device */
       int minor;
};

/* linked list of alarms to check for */
static struct gpio_private *alarmlist = 0;

#if defined(CONFIG_ARGUS_MEMORY_MAPPED_GPIO)
unsigned* mm_gpio_addr = 0; 
unsigned mm_gpio_shadow = 0;
#endif

/* Helper functions */
static inline unsigned long
gpio_read_bits(struct gpio_private *priv)
{
	int bit;
	unsigned long int val = 0;
	for (bit = 0; bit < PORT_BITS; bit++) 
		val |= gpio_get_level(priv->base + bit) << bit;	    
	return val;
}

static inline unsigned long
gpio_check_trig(struct gpio_private *priv)
{
	int bit;
	unsigned long int val = 0;
	for (bit = 0; bit < PORT_BITS; bit++) 
		val |= gpio_get_trig(priv->base + bit) << bit;	    
	return val;
}


static inline void
gpio_write_bits(struct gpio_private *priv, unsigned long data, unsigned long mask)
{
	int bit;
	for (bit = 0; bit < PORT_BITS; bit++) {
		if (mask & (1 << bit))
			gpio_set_mode(priv->base + bit, GPIO_MODE_OUT, (data & (1<<bit))>>bit);   		
	}
}

static inline void
gpio_set_dir(struct gpio_private *priv, unsigned long mask, unsigned long dir)
{
	int bit;
	for (bit = 0; bit < PORT_BITS; bit++) {
		if (mask & (1 << bit))
			gpio_set_mode(priv->base + bit, dir, 0);
	}
}

static inline unsigned long
gpio_get_dir(struct gpio_private *priv)
{
	int bit;
	unsigned long ret = 0;
	for (bit = 0; bit < PORT_BITS; bit++) {
		if (GET_GPIO_MODE(SOC_GPIO.PIN[priv->base + bit]) == GPIO_MODE_OUT)
			ret |= (1 << bit);
	}
	return ret;
}

static inline void
gpio_enable_irqs(struct gpio_private *priv, unsigned long mask, int polarity)
{
	int bit;
	for (bit = 0; bit < PORT_BITS; bit++) {
		if (mask & (1 << bit)) { 
			gpio_set_edge_detect_polarity(priv->base + bit, polarity);
			gpio_clear_trig(priv->base + bit);
			gpio_set_mode(priv->base + bit, GPIO_MODE_IN_IRQ, 0);
		}
	}	
}

static void
gpio_clralarm(struct gpio_private *priv, unsigned long mask)
{
	int bit;
	struct gpio_private *p;
	unsigned long lowalarm = 0, highalarm = 0;

	/* Check what alarms are left */
	p = alarmlist;
	while (p) {
		/* Only use masks for users of this port. */
		if (p->base == priv->base) {	
			highalarm |= p->highalarm;
			lowalarm |= p->lowalarm;
		}
		p = p->next;
	}
	/* Clean up the pins with alarms where there are no other alarms */
	for (bit = 0; bit < PORT_BITS; bit++) {
		if ((1 << bit) & (~(highalarm|lowalarm)) & mask) {
			int mode;
			gpio_clear_trig(priv->base + bit);
			mode = GET_GPIO_MODE(SOC_GPIO.PIN[priv->base + bit]);
			switch (mode) {
				case GPIO_MODE_IN_IRQ:
				case GPIO_MODE_IN_FIQ:
				  mode = GPIO_MODE_IN;
				  break;
				case GPIO_MODE_IN_PU_IRQ:
				case GPIO_MODE_IN_PU_FIQ:
				  mode = GPIO_MODE_IN_PU;
				  break;
			}
			gpio_set_mode(priv->base + bit, mode, 0);
			D(printk("gpio_clralarm clearing bit %i+%i\n", priv->base, bit));
		}
	}

}

static unsigned int
gpio_poll(struct file *file, poll_table *wait)
{
	struct gpio_private *priv = (struct gpio_private *)file->private_data;
	unsigned int val;
	int bit;

	if (priv->minor > GPIO_MINOR_LAST_PWM ||
	    (priv->minor > GPIO_MINOR_LAST_GPIO &&
	     priv->minor < GPIO_MINOR_PWM0) ||
	    priv->minor == GPIO_MINOR_LEDS ||
	    priv->minor == GPIO_MINOR_PWM0)
		return -EINVAL;

	poll_wait(file, &priv->alarm_wq, wait);
	val = gpio_read_bits(priv);
	D(printk("gpio_poll: val %lx  bit1: %i high %lx low %lx  H %lx L %x\n", (unsigned long)val, val & 2, priv->highalarm, priv->lowalarm, val & priv->highalarm, (unsigned long)(~val & priv->lowalarm)));

	/* The gpios support edge interrupt -
	 * lets activate posedge for those low and with highalarm set and
	 * negedge for those high and with lowalarm set
	 */
	for (bit = 0; bit < PORT_BITS; bit++) {
		if ((~val & priv->highalarm & (1 << bit))) { 
			/* posedge */
			gpio_set_edge_detect_polarity(priv->base + bit, 1);
			gpio_clear_trig(priv->base + bit);
			gpio_set_mode(priv->base + bit, GPIO_MODE_IN_IRQ, 0);
		} else if ((val & priv->lowalarm & (1 << bit))) { 
			/* negedge */
			gpio_set_edge_detect_polarity(priv->base + bit, 0);
			gpio_clear_trig(priv->base + bit);
			gpio_set_mode(priv->base + bit, GPIO_MODE_IN_IRQ, 0);
		}
	}		

	/* Unmask GPIO interrupt */
	gpio_irq_enable();

	/* Check if alarm is active */
	if ((val & priv->highalarm) ||
	    (~val & priv->lowalarm)) {
		return POLLIN|POLLRDNORM;
	}	
	return 0;
}

static irqreturn_t
gpio_interrupt(int irq, void *dev_id)
{
	struct gpio_private *priv = alarmlist;
	unsigned long int trigs = 0;
	int bit;
	unsigned int irq_handled = 0;
	static unsigned long prev_ack_jiffies = 0;
	static long unacked_irq_count = 0;
	
	DINT(printk("gpio_int %i PRIO: IRQ_NBR: %i IRQ: %i FIQ_NBR %i FIQ %i level %i pin %02x\n",
		 irq, 
		 GET_GPIO_PRIO_IRQ_NBR(SOC_GPIO.PRIO),
		 GET_GPIO_PRIO_IRQ(SOC_GPIO.PRIO),
		 GET_GPIO_PRIO_FIQ_NBR(SOC_GPIO.PRIO),
		 GET_GPIO_PRIO_FIQ(SOC_GPIO.PRIO),
		 gpio_get_level(GET_GPIO_PRIO_IRQ_NBR(SOC_GPIO.PRIO)), SOC_GPIO.PIN[GET_GPIO_PRIO_IRQ_NBR(SOC_GPIO.PRIO)])
		 ); 

	unacked_irq_count++;

	/* Look for pending alarms */
	while (priv) {
		if (priv->highalarm || priv->lowalarm) {
			trigs = gpio_check_trig(priv);

			if (trigs & (priv->highalarm | priv->lowalarm)) {
				wake_up_interruptible(&priv->alarm_wq);
			}
			/* ACK IRQs */
			for (bit = 0; bit < PORT_BITS; bit++) {
				if (trigs & (1 << bit) & (priv->highalarm | priv->lowalarm)) {
					gpio_clear_trig(priv->base + bit);
					unacked_irq_count = 0;
					prev_ack_jiffies = jiffies;
					irq_handled = 1;
				}
			}
		}
		priv = priv->next;
	}

#if 0
	if (unacked_irq_count > 10) {
		if (GET_GPIO_PRIO_IRQ(SOC_GPIO.PRIO)) {
			gpio_clear_trig(GET_GPIO_PRIO_IRQ_NBR(SOC_GPIO.PRIO));
			D(printk("gpio_int # many unacked!\n"));
		}
	}
#endif

	return IRQ_RETVAL(irq_handled);
}

static int
gpio_open(struct inode *inode, struct file *filp)
{
	struct gpio_private* priv;
	int minor =  MINOR(inode->i_rdev);

	D(printk("GPIO open %d\n", minor));	

#if defined(CONFIG_ARGUS_MEMORY_MAPPED_GPIO)
	if ((minor > GPIO_MINOR_LAST_PWM
	     || (minor > GPIO_MINOR_LAST_GPIO && minor < GPIO_MINOR_PWM0))
	    && minor != GPIO_MINOR_H)
		return -EINVAL;
#else
	if (minor > GPIO_MINOR_LAST_PWM ||
	    (minor > GPIO_MINOR_LAST_GPIO &&
	     minor < GPIO_MINOR_PWM0))
		return -EINVAL;
#endif

	priv = (struct gpio_private*)kmalloc(sizeof *priv, GFP_KERNEL);
	if (!priv)
		return -ENOMEM;
        
	priv->minor = minor;
	filp->private_data = (void *)priv;

	// Only setup alarmlists etc. for the actual GPIO ports (not the PWM ports)
	if(minor <= GPIO_MINOR_LAST_GPIO && minor != GPIO_MINOR_LEDS) {
		priv->next = alarmlist;
		priv->highalarm = priv->lowalarm = 0;
		init_waitqueue_head(&priv->alarm_wq);
		if (minor == 0) {
			priv->base = 0;
			priv->changeable_bits = CONFIG_ARGUS_GPIO_CHANGEABLE_BITS1;
		} else if (minor == 1) {
			priv->base = PORT_BITS;
			priv->changeable_bits = CONFIG_ARGUS_GPIO_CHANGEABLE_BITS2;
		} else if (minor == 3) {
			priv->base = PORT_BITS * 2;
			priv->changeable_bits = CONFIG_ARGUS_GPIO_CHANGEABLE_BITS3;
		}
		/* Insert in list when all initialised */
		alarmlist = priv;
	}

	return 0;
}

static int
gpio_release(struct inode *inode, struct file *filp)
{
	struct gpio_private *p = alarmlist;
	struct gpio_private *todel = (struct gpio_private *)filp->private_data;
	unsigned int flags;
	
	D(printk("GPIO release %d\n", MINOR(inode->i_rdev)));
	// unlink from alarmlist and free the private structure
	// Only do this for the real GPIO ports

	if(todel->minor <= GPIO_MINOR_LAST_GPIO && todel->minor != GPIO_MINOR_LEDS) {
		gpio_irq_save_disable(flags);
		if (p == todel) {
			alarmlist = todel->next;
		} else {
			while (p->next != todel)
				p = p->next;
			p->next = todel->next;
		}
		gpio_clralarm(todel,  todel->highalarm | todel->lowalarm);
		gpio_irq_restore(flags);
	}

	kfree(todel);
	return 0;
}

static int gpio_leds_ioctl(unsigned int cmd, unsigned long arg);
static int gpio_pwm_ioctl(struct gpio_private *priv, 
			  unsigned int cmd, unsigned long arg);
#if defined(CONFIG_ARGUS_MEMORY_MAPPED_GPIO)
static int gpio_memory_mapped_ioctl(unsigned int cmd, unsigned long arg);
#endif

static int 
gpio_ioctl(struct inode* inode, struct file* file,
           unsigned int cmd, unsigned long arg)
{
	int val = 0;
	unsigned int flags;

	struct gpio_private* priv = (struct gpio_private*)file->private_data;

	D(printk("GPIO IOCTL cmd %x arg %X\n", cmd, arg));

#if defined(CONFIG_ARGUS_MEMORY_MAPPED_GPIO)
	if (priv->minor == GPIO_MINOR_H)
		return gpio_memory_mapped_ioctl(cmd, arg);
#endif

	// Check for other ioctl handlers
	if (priv->minor == GPIO_MINOR_LEDS)
		return gpio_leds_ioctl(cmd, arg);
	else if (priv->minor >= GPIO_MINOR_PWM0 &&
		 priv->minor <= GPIO_MINOR_LAST_PWM)
		return gpio_pwm_ioctl(priv, cmd, arg);

	switch(_IOC_NR(cmd)) {
	case IO_READBITS: /* Use IO_READ_INBITS and IO_READ_OUTBITS instead */
		val = gpio_read_bits(priv);
		D(printk("IO_READBITS: %08lx gpio1: %i\n", (unsigned long)val, val & 2));
		return val;
	case IO_SETBITS:
		gpio_write_bits(priv, 0xffffffff, arg & priv->changeable_bits);
		break;
	case IO_CLRBITS:
		gpio_write_bits(priv, 0, arg & priv->changeable_bits);
		break;
	case IO_HIGHALARM:	
		gpio_irq_save_disable(flags);
		// set alarm when bits with 1 in arg go high
		priv->highalarm |= arg;
		D(printk("HIGHALARM H %x L %x\n", priv->highalarm, priv->lowalarm));
		/* irqs are enabled in gpio_poll */
		//gpio_enable_irqs(priv, priv->highalarm, 1);
		gpio_irq_restore(flags);
		break;
	case IO_LOWALARM:
		gpio_irq_save_disable(flags);
		// set alarm when bits with 1 in arg go low
		priv->lowalarm |= arg;
		D(printk("LOWALARM  H %x L %x\n", priv->highalarm, priv->lowalarm));
		/* irqs are enabled in gpio_poll */
		//gpio_enable_irqs(priv, priv->lowalarm, 0);
		gpio_irq_restore(flags);
		break;	
	case IO_CLRALARM:
		// clear alarm for bits with 1 in arg
		gpio_irq_save_disable(flags);
		priv->highalarm &= ~arg;
		priv->lowalarm  &= ~arg;
		D(printk("CLRALARM  H %x L %x\n", priv->highalarm, priv->lowalarm));
		gpio_clralarm(priv, arg);
		gpio_irq_restore(flags);
		break;	
	case IO_READDIR: /* Use IO_SETGET_INPUT/OUTPUT instead! */
		gpio_get_dir(priv);
		break;	 
	case IO_SETINPUT: /* Use IO_SETGET_INPUT instead! */
		gpio_set_dir(priv, arg & priv->changeable_bits, GPIO_MODE_IN);
		return gpio_read_bits(priv) & ~gpio_get_dir(priv);
	case IO_SETOUTPUT: /* Use IO_SETGET_OUTPUT instead! */
		gpio_set_dir(priv, arg & priv->changeable_bits, GPIO_MODE_OUT);
		return gpio_read_bits(priv) & gpio_get_dir(priv);
	case IO_READ_INBITS: 
		if (copy_from_user(&val, (unsigned long*)arg, sizeof(val)))
			return -EFAULT;
		/* *arg is result of reading the input/output pins */
		val = gpio_read_bits(priv) & ~gpio_get_dir(priv);
		if (copy_to_user((unsigned long*)arg, &val, sizeof(val)))
			return -EFAULT;
		break;
	case IO_READ_OUTBITS:
		if (copy_from_user(&val, (unsigned long*)arg, sizeof(val)))
			return -EFAULT;
		/* *arg is result of reading the input/output pins */
		val = gpio_read_bits(priv) & gpio_get_dir(priv);
		if (copy_to_user((unsigned long*)arg, &val, sizeof(val)))
			return -EFAULT;
		break;
	case IO_SETGET_INPUT: 
		/* bits set in *arg is set to input,
		 * *arg updated with current input pins.
		 */
		if (copy_from_user(&val, (unsigned long*)arg, sizeof(val)))
			return -EFAULT;
		gpio_set_dir(priv, val & priv->changeable_bits, GPIO_MODE_IN);
		val = ~gpio_get_dir(priv);
		if (copy_to_user((unsigned long*)arg, &val, sizeof(val)))
			return -EFAULT;
		break;
	case IO_SETGET_OUTPUT:
		/* bits set in *arg is set to output,
		 * *arg updated with current output pins.
		 */
		if (copy_from_user(&val, (unsigned long*)arg, sizeof(val)))
			return -EFAULT;
		gpio_set_dir(priv, val & priv->changeable_bits, GPIO_MODE_OUT);
		val = gpio_get_dir(priv);
		if (copy_to_user((unsigned long*)arg, &val, sizeof(val)))
			return -EFAULT;
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

#if defined(CONFIG_ARGUS_MEMORY_MAPPED_GPIO)
static int
gpio_memory_mapped_ioctl(unsigned int cmd, unsigned long arg)
{
	int val = 0;

	D(printk("Memory-mapped GPIO IOCTL cmd %x arg %X\n", cmd, arg));
	
	switch (_IOC_NR(cmd)) {
	case IO_SETBITS:	   
		mm_gpio_shadow |= (arg & CONFIG_ARGUS_MEMORY_MAPPED_GPIO_CHANGEABLE_BITS);
		*mm_gpio_addr = mm_gpio_shadow << 16;
		break;
	case IO_CLRBITS:
		mm_gpio_shadow &= ~(arg & CONFIG_ARGUS_MEMORY_MAPPED_GPIO_CHANGEABLE_BITS);
		*mm_gpio_addr = mm_gpio_shadow << 16;
		break;
	case IO_READ_OUTBITS:
		if (copy_from_user(&val, (unsigned long*)arg, sizeof(val)))
			return -EFAULT;
		val = mm_gpio_shadow;
		if (copy_to_user((unsigned long*)arg, &val, sizeof(val)))
			return -EFAULT;
		break;
	case IO_SETGET_OUTPUT:
		if (copy_from_user(&val, (unsigned long*)arg, sizeof(val)))
			return -EFAULT;
		val = 0xFFFFFFFF;
		if (copy_to_user((unsigned long*)arg, &val, sizeof(val)))
			return -EFAULT;
		break;
	default:
		return -EINVAL;
	}

	return 0;
}
#endif

// Handle the LEDS ports ioctls

static int
gpio_leds_ioctl(unsigned int cmd, unsigned long arg)
{
	unsigned char green;
	unsigned char red;

	switch (_IOC_NR(cmd)) {
	case IO_LEDACTIVE_SET:
		green = ((unsigned char) arg) & 1;
		red   = (((unsigned char) arg) >> 1) & 1;
		LED_ACTIVE_SET_G(green);
		LED_ACTIVE_SET_R(red);
		break;
	default:
		return -EINVAL;
	} /* switch */

	return 0;
}

// Handle the PWM-ports ioctls

static int
gpio_pwm_ioctl(struct gpio_private *priv, unsigned int cmd, unsigned long arg)
{
	int pwmport = priv->minor - GPIO_MINOR_PWM0;

	switch (_IOC_NR(cmd)) {
		case IO_PWM_SET_PRESCALE:
			SOC_PWM[pwmport].PRESCALE = arg & 0x3;
			// Make sure the pin is set to ALT_FUNC
			gpio_set_mode(pwmport ? GPIO_PWM1 : GPIO_PWM0,
				      GPIO_MODE_ALT_FUNC, 0);
			break;
		case IO_PWM_SET_PERIOD:
			SOC_PWM[pwmport].PERIOD = arg;
			break;
		default:
			return -EINVAL;
	}

	return 0;
}

struct file_operations gpio_fops = {
	owner:       THIS_MODULE,
	poll:        gpio_poll,
	ioctl:       gpio_ioctl,
	open:        gpio_open,
	release:     gpio_release,
};

/* Driver initialization */
static __init int
gpio_init(void)
{
	int res;
  
	res = register_chrdev(GPIO_MAJOR, "argus gpio" , &gpio_fops);
  	if (res < 0) {
		printk(KERN_ERR "gpio: couldn't get a major number.\n");
		return res;
	}

	printk("Argus GPIO driver\n");
 
#if defined(CONFIG_ARGUS_MEMORY_MAPPED_GPIO)
	mm_gpio_addr = (unsigned*)ioremap_nocache(EXT2_BASE_P, sizeof(*mm_gpio_addr));
	mm_gpio_shadow = CONFIG_ARGUS_MEMORY_MAPPED_GPIO_DEFAULT_VALUES;
	*mm_gpio_addr = mm_gpio_shadow << 16;
#endif

	gpio_irq_disable();
	if (request_irq(IRQ_GPIO_IRQ, gpio_interrupt, IRQF_SHARED ,"gpio", &alarmlist)) {
		printk("err: irq for gpio\n");
	}
	return res;

}

module_init(gpio_init);

#if defined(CONFIG_ARGUS_MEMORY_MAPPED_GPIO)
EXPORT_SYMBOL(mm_gpio_addr);
EXPORT_SYMBOL(mm_gpio_shadow);
#endif
