/*
 * Ethernet driver for Open Ethernet Controller (www.opencores.org).
 *      Copyright (c) 2002 Simon Srot (simons@opencores.org)
 *
 * Changes:
 * 10. 02. 2004: Matjaz Breskvar (phoenix@opencores.org)
 *   port to linux-2.4, workaround for next tx bd number. 
 *
 *      Rewritten 2006 for Argus-4
 *          Copyright (c) 2006 Axis Communications AB, Sweden
 *
 * Based on:
 *
 * Ethernet driver for Motorola MPC8xx.
 *      Copyright (c) 1997 Dan Malek (dmalek@jlc.net)
 *
 * mcen302.c: A Linux network driver for Mototrola 68EN302 MCU
 *
 *      Copyright (C) 1999 Aplio S.A. Written by Vadim Lebedev
 *
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/ptrace.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/spinlock.h>
#include <linux/crc32.h>
#include <linux/mii.h>
#include <linux/ethtool.h>

#include <asm/irq.h>
#include <asm/pgtable.h>
#include <asm/bitops.h>
#include <asm/uaccess.h>
#include <asm/cacheflush.h>
#include <asm/arch/gpio.h>

/* __PHX__ :: clean up :: for ETH_BASE_ADDR and MAXADDR? */

#include "open_eth.h"

#define __clear_user(add,len) memset((add),0,(len))

//#define DEBUG

#define _print printk

#ifdef DEBUG
#define D(x) x
#else
#define D(x)
#endif

#define DMDIO(x)

/* The transmitter timeout
 */
#define TX_TIMEOUT	(2*HZ)

/* Buffer number (must be 2^n) 
 */
#define OETH_RXBD_NUM		8
#define OETH_TXBD_NUM		8
#define OETH_RXBD_NUM_MASK	(OETH_RXBD_NUM-1)
#define OETH_TXBD_NUM_MASK	(OETH_TXBD_NUM-1)

/* Buffer size 
 */
#define OETH_RX_BUFF_SIZE	2048
#define OETH_TX_BUFF_SIZE	2048

/* How many buffers per page 
 */
#define OETH_RX_BUFF_PPGAE	(PAGE_SIZE/OETH_RX_BUFF_SIZE)
#define OETH_TX_BUFF_PPGAE	(PAGE_SIZE/OETH_TX_BUFF_SIZE)

/* How many pages are needed for buffers 
 */
#define OETH_RX_BUFF_PAGE_NUM	(OETH_RXBD_NUM/OETH_RX_BUFF_PPGAE)
#define OETH_TX_BUFF_PAGE_NUM	(OETH_TXBD_NUM/OETH_TX_BUFF_PPGAE)

// Buffer size, aligned to 4-bytes
// This has to include the max framesize + NET_IP_ALIGN
#define MAX_FRAME_SIZE		1520
// Taken out of the thin air.. breakeven between copying a packet and wasting
// memory by sending up the entire 1520 byte buffer.
#define RX_COPYBREAK 128

/* The buffer descriptors track the ring buffers.   
 */
struct oeth_private {
	struct	sk_buff* rx_skbuff[OETH_RXBD_NUM];
	struct	sk_buff* tx_skbuff[OETH_TXBD_NUM];

	ushort	tx_next;			/* Next buffer to be sent */
	ushort	tx_last;			/* Next buffer to be checked if packet sent */
	ushort	tx_full;			/* Buffer ring fuul indicator */
	ushort	rx_cur;				/* Next buffer to be checked if packet received */

	oeth_regs	*regs;			/* Address of controller registers. */
	oeth_bd		*rx_bd_base;		/* Address of Rx BDs. */
	oeth_bd		*tx_bd_base;		/* Address of Tx BDs. */

	struct net_device_stats stats;
};

// Duplex settings
enum duplex
{
	half,
	full,
	autoneg
};

static int oeth_open(struct net_device *dev);
static int oeth_start_xmit(struct sk_buff *skb, struct net_device *dev);
static void oeth_rx(struct net_device *dev);
static void oeth_tx(struct net_device *dev);
static irqreturn_t oeth_interrupt(int irq, void *dev_id);
static int oeth_close(struct net_device *dev);
static struct net_device_stats *oeth_get_stats(struct net_device *dev);
static void oeth_set_multicast_list(struct net_device *dev);
static int oeth_set_mac_add(struct net_device *dev, void *addr);
static int calc_crc(char *mac_addr);

static void clear_network_leds(unsigned long dummy);
static void set_network_leds(int active);
static void check_speed(unsigned long dummy);
static void set_speed(struct net_device* dev, unsigned long speed);
static void check_duplex(unsigned long dummy);
static void set_duplex(struct net_device* dev, enum duplex);

// Transceiver specific funtions
static void tdk_check_speed(struct net_device* dev);
static void tdk_check_duplex(struct net_device* dev);
static void bmcr_check_speed(struct net_device* dev);
static void bmcr_check_duplex(struct net_device* dev);
static void natsemi_check_speed(struct net_device* dev);
static void natsemi_check_duplex(struct net_device* dev);
//static void davicom_check_speed(struct net_device* dev);
//static void davicom_check_duplex(struct net_device* dev);
static void generic_check_speed(struct net_device* dev);
static void generic_check_duplex(struct net_device* dev);

struct transceiver_ops
{
	unsigned int oui;
	void (*check_speed)(struct net_device* dev);
	void (*check_duplex)(struct net_device* dev);
};

struct transceiver_ops transceivers[] = 
{
	{0xC039, tdk_check_speed, tdk_check_duplex},            /* TDK 2120 */
	{0x039C, tdk_check_speed, tdk_check_duplex},            /* TDK 2120C */
	{0x0020, bmcr_check_speed, bmcr_check_duplex}, /* RealTek 8201 (RealTek cheats with OUI to MII PHY id mapping) */
	{0x80017, natsemi_check_speed, natsemi_check_duplex },  /* National Semiconductors DP83846A. */
	{0x606E, bmcr_check_speed, bmcr_check_duplex },  /* Davicom DM9161 */
	{0x0000, generic_check_speed, generic_check_duplex}     /* Generic, must be last */
};

#ifdef DEBUG
static void
oeth_print_packet(unsigned long add, int len)
{
	int i;

	_print("ipacket: add = %lx len = %d\n", add, len);
	_print("DADDR: %02x%02x%02x%02x%02x%02x, ", *(((unsigned char *)add)), *(((unsigned char *)add) + 1), *(((unsigned char *)add) + 2), *(((unsigned char *)add) + 3), *(((unsigned char *)add) + 4), *(((unsigned char *)add) + 5));
	_print("SADDR: %02x%02x%02x%02x%02x%02x, ", *(((unsigned char *)add)+6), *(((unsigned char *)add) + 7), *(((unsigned char *)add) + 8), *(((unsigned char *)add) + 9), *(((unsigned char *)add) + 10), *(((unsigned char *)add) + 11));
	_print("Type/len: %02x%02x\n", *(((unsigned char *)add)+12), *(((unsigned char *)add) + 13));
	for(i = 14; i < len+16*4; i++) {
  		if(i == len)
   			_print("***\n");
  		if(!(i % 16))
    			_print("\n");
  		_print(" %.2x", *(((unsigned char *)add) + i));
	}
	_print("\n");
}
#endif

// MAC clock management.
// TODO.
static atomic_t mac_clock_users;
#if 0
#define ETH_CLOCK_START do {if (atomic_add_return(1, &mac_clock_users) == 1) SOC_CLK.ENABLE = CLK_ETH; } while (0)
#define ETH_CLOCK_STOP  do {if (atomic_add_return(-1, &mac_clock_users) == 0) SOC_CLK.DISABLE = CLK_ETH; } while (0)
#else
#define ETH_CLOCK_START
#define ETH_CLOCK_STOP
#endif

struct transceiver_ops* transceiver;

static unsigned int mdio_phy_addr; /* Transceiver address. */

/* Network speed indication. */
static struct timer_list speed_timer = TIMER_INITIALIZER(NULL, 0, 0);
static struct timer_list clear_led_timer = TIMER_INITIALIZER(NULL, 0, 0);
static int current_speed; /* Speed read from tranceiver */
static int current_speed_selection; /* Speed selected by user */
static unsigned long led_next_time;
static int led_active;

/* Duplex */
static struct timer_list duplex_timer = TIMER_INITIALIZER(NULL, 0, 0);
static int full_duplex;
static enum duplex current_duplex;

/* Broadcom specific */
#define MDIO_AUX_CTRL_STATUS_REG           0x18
#define MDIO_BC_FULL_DUPLEX_IND             0x1
#define MDIO_BC_SPEED                       0x2

/* TDK specific */
#define MDIO_TDK_DIAGNOSTIC_REG              18
#define MDIO_TDK_DIAGNOSTIC_RATE          0x400
#define MDIO_TDK_DIAGNOSTIC_DPLX          0x800

/* National Semiconductors DP83846A specific */
#define MDIO_NATSEMI_PHYSTS_REG             0x10
#define MDIO_NATSEMI_PHYSTS_LINK          (1<<0)
#define MDIO_NATSEMI_PHYSTS_SPEED         (1<<1)
#define MDIO_NATSEMI_PHYSTS_DPLX          (1<<2)

#if 0
// Davicom DM9161 specific 
#define MDIO_DAVICOM_DSCSR_REG                17
#define MDIO_DAVICOM_DSCSR_100FDX        (1<<15)
#define MDIO_DAVICOM_DSCSR_100HDX        (1<<14)
#define MDIO_DAVICOM_DSCSR_10FDX         (1<<13)
#define MDIO_DAVICOM_DSCSR_10HDX         (1<<12)
#endif

/* Network flash constants */
#define NET_FLASH_TIME                  (HZ/50) /* 20 ms */
#define NET_FLASH_PAUSE                (HZ/100) /* 10 ms */
#define NET_LINK_UP_CHECK_INTERVAL        (2*HZ) /* 2 s   */
#define NET_DUPLEX_CHECK_INTERVAL         (2*HZ) /* 2 s   */

#define NO_NETWORK_ACTIVITY 0
#define NETWORK_ACTIVITY    1

// Support routines for handling MII communications with the PHY.
// 
// The MAC handles the serial protocol, we just need to write to
// some registers and wait for completion.
//
// TODO: The write routine does not wait for completion - this
// could upset code that relies on synchronous updates of the PHY
// state (check).

static void 
set_mdio_reg(int addr, int data)
{
	DMDIO(printk("set_mdio_reg: addr=%x data=%x\n", addr, data));
	// Wait for BUSY flag to deassert
	while(GET_MIISTATUS_BUSY(SOC_ETHERNET.MIISTATUS) == 1)
		;
	SOC_ETHERNET.MIIADDRESS =
		SET_MIIADDRESS_FIAD(mdio_phy_addr) |
		SET_MIIADDRESS_RGAD(addr); 
	SOC_ETHERNET.MIITX_DATA = SET_MIITX_DATA_CTRLDATA(data);
	SOC_ETHERNET.MIICOMMAND = OETH_MIICOMMAND_WCTRLDATA;
}

static int
get_mdio_reg(int addr)
{
	int data;
	// Wait for BUSY flag to deassert
	while(GET_MIISTATUS_BUSY(SOC_ETHERNET.MIISTATUS) == 1)
		;
	SOC_ETHERNET.MIIADDRESS =
		SET_MIIADDRESS_FIAD(mdio_phy_addr) |
		SET_MIIADDRESS_RGAD(addr); 

	SOC_ETHERNET.MIICOMMAND = OETH_MIICOMMAND_RSTAT;
	
	// Wait for BUSY flag to deassert again
	while(GET_MIISTATUS_BUSY(SOC_ETHERNET.MIISTATUS) == 1)
		;
 	data = GET_MIIRX_DATA_PRSD(SOC_ETHERNET.MIIRX_DATA);
 	//DMDIO(printk("get_mdio_reg: addr=%x data=%x\n", addr, data));
	return data;
}

// Dump all mdio regs from 0 to 31

static void
dump_mdio_regs(void)
{
	int i;

	printk("Dumping all MDIO regs from PHY with addr %d.\n",
	       mdio_phy_addr);

	for(i = 0; i < 32; i++) {
		printk("Reg %d: 0x%08x\n", i, get_mdio_reg(i));
	}
}

static void
phy_negotiate(struct net_device *dev)
{
	unsigned short data = get_mdio_reg(MII_ADVERTISE);

	/* Discard old speed and duplex settings */
	data &= ~(ADVERTISE_100HALF | ADVERTISE_100FULL | 
		  ADVERTISE_10FULL | ADVERTISE_10HALF);
  
	switch (current_speed_selection) {
		case 10 :
			if (current_duplex == full)
				data |= ADVERTISE_10FULL;
			else if (current_duplex == half)
				data |= ADVERTISE_10HALF;
			else
				data |= ADVERTISE_10HALF | ADVERTISE_10FULL;
			break;

		case 100 :
			 if (current_duplex == full)
				data |= ADVERTISE_100FULL;
			else if (current_duplex == half)
				data |= ADVERTISE_100HALF;
			else
				data |= ADVERTISE_100HALF | ADVERTISE_100FULL;
			break;

		case 0 : /* Auto */
			 if (current_duplex == full)
				data |= ADVERTISE_100FULL | ADVERTISE_10FULL;
			else if (current_duplex == half)
				data |= ADVERTISE_100HALF | ADVERTISE_10HALF;
			else
				data |= ADVERTISE_100HALF | ADVERTISE_100FULL | 
					ADVERTISE_10FULL | ADVERTISE_10HALF;
			break;

		default : /* assume autoneg speed and duplex */
			data |= ADVERTISE_100HALF | ADVERTISE_100FULL | 
				ADVERTISE_10FULL | ADVERTISE_10HALF;

	}

	set_mdio_reg(MII_ADVERTISE, data);
	
	/* Renegotiate with link partner */
	data = get_mdio_reg(MII_BMCR);
	data |= BMCR_ANRESTART;	
	set_mdio_reg(MII_BMCR, data);
}

static void 
phy_reset(void)
{
	int readval,i;
	readval = get_mdio_reg(MII_BMCR);
	readval |= BMCR_RESET;
	set_mdio_reg(MII_BMCR, readval);
	i = 0;
	while(get_mdio_reg(MII_BMCR) & BMCR_RESET)
		if (++i > 100) {
			printk("ERROR: resetting PHY");
			break;
		}
}

static void
generic_check_speed(struct net_device *dev)
{
	unsigned long data;
	data = get_mdio_reg(MII_ADVERTISE);
	if ((data & ADVERTISE_100FULL) ||
	    (data & ADVERTISE_100HALF))
		current_speed = 100;
	else
		current_speed = 10;
}

static void
tdk_check_speed(struct net_device *dev)
{
	unsigned long data;
	data = get_mdio_reg(MDIO_TDK_DIAGNOSTIC_REG);
	current_speed = (data & MDIO_TDK_DIAGNOSTIC_RATE ? 100 : 10);
}

static void
bmcr_check_speed(struct net_device *dev)
{
	unsigned long data;
	data = get_mdio_reg(MII_BMCR);
	current_speed = (data & BMCR_SPEED100) ? 100 : 10;
}

static void
natsemi_check_speed(struct net_device *dev)
{
	unsigned long data;
	data = get_mdio_reg(MDIO_NATSEMI_PHYSTS_REG);
	current_speed = (data & MDIO_NATSEMI_PHYSTS_SPEED) ? 10 : 100;
}

#if 0
static void
davicom_check_speed(struct net_device *dev)
{
	unsigned long data;
	// Read DAVICOM Specified Configuration and Status Register
	data = get_mdio_reg(dickdock);
	// If it is in 100FDX or 100HDX, we assume 100 mbit op,
	// otherwise 10
	current_speed = (data & (MDIO_DAVICOM_DSCSR_100FDX |
				 MDIO_DAVICOM_DSCSR_100HDX )) ? 100 : 10;
}
#endif

static void
check_speed(unsigned long d)
{
	unsigned long data;
	int old_speed;

	ETH_CLOCK_START;

	old_speed = current_speed;

	data = get_mdio_reg(MII_BMSR);
		
	if (!(data & BMSR_LSTATUS)) {
		current_speed = 0;
	} else {
		transceiver->check_speed((struct net_device*)d);
	}
	
	if (old_speed != current_speed) {
		if (current_speed)
			netif_carrier_on((struct net_device*)d);
		else
			netif_carrier_off((struct net_device*)d);
		
		set_network_leds(NO_NETWORK_ACTIVITY);
	}

	ETH_CLOCK_STOP;

	/* Reinitialize the timer. */
	speed_timer.expires = jiffies + NET_LINK_UP_CHECK_INTERVAL;
	add_timer(&speed_timer);
}

static void
set_speed(struct net_device* dev, unsigned long speed)
{
	if (current_speed_selection != speed) {
		current_speed_selection = speed;
		phy_negotiate(dev);
	}
}

static void
generic_check_duplex(struct net_device* dev)
{
	unsigned long data;
	data = get_mdio_reg(MII_ADVERTISE);
	if ((data & ADVERTISE_10FULL) ||
	    (data & ADVERTISE_100FULL))
		full_duplex = 1;
	else
		full_duplex = 0;
}

static void
tdk_check_duplex(struct net_device* dev)
{
	unsigned long data;
	data = get_mdio_reg(MDIO_TDK_DIAGNOSTIC_REG);
	full_duplex = (data & MDIO_TDK_DIAGNOSTIC_DPLX) ? 1 : 0;
}

static void
bmcr_check_duplex(struct net_device* dev)
{
	unsigned long data;
	data = get_mdio_reg(MII_BMCR);
	full_duplex = (data & BMCR_FULLDPLX) ? 1 : 0;
}

static void
natsemi_check_duplex(struct net_device *dev)
{
	unsigned long data;
	data = get_mdio_reg(MDIO_NATSEMI_PHYSTS_REG);
	full_duplex = (data & MDIO_NATSEMI_PHYSTS_DPLX) ? 1 : 0;
}

#if 0
static void
davicom_check_duplex(struct net_device *dev)
{
	unsigned long data;

	// Read DAVICOM Specified Configuration and Status Register
	data = get_mdio_reg(MDIO_DAVICOM_DSCSR_REG);
	// If it is in 100FDX or 10FDX, we assume full duplex op
	// otherwise half
	full_duplex = (data & (MDIO_DAVICOM_DSCSR_100FDX |
			       MDIO_DAVICOM_DSCSR_10FDX )) ? 1 : 0;
}
#endif

static void
check_duplex(unsigned long d)
{
	int old_duplex = full_duplex;
#if 0
	ETH_CLOCK_START;
	transceiver->check_duplex((struct net_device*)d);
	if (old_duplex != full_duplex) {
		DMDIO(printk("Changing duplex to %d\n", full_duplex));
		if (full_duplex) {
			/* Duplex changed to full. */
			SOC_ETHERNET.MODER |= OETH_MODER_FULLD; 
		} else {
			/* Duplex changed to half. */
			SOC_ETHERNET.MODER &= ~OETH_MODER_FULLD;
		}		
	}
	ETH_CLOCK_STOP;
#endif
	/* Reinitialize the timer. */
	duplex_timer.expires = jiffies + NET_DUPLEX_CHECK_INTERVAL;
	add_timer(&duplex_timer);
}

static int 
probe_transceiver(struct net_device* dev)
{
	unsigned int phyid_high;
	unsigned int phyid_low;
	unsigned int oui;
	struct transceiver_ops* ops = NULL;

	/* Probe MDIO physical address */
	for (mdio_phy_addr = 0; mdio_phy_addr <= 31; mdio_phy_addr++) {
		if (get_mdio_reg(MII_BMSR) != 0xffff)
			break;
	}
	printk(KERN_INFO "PHY MII address %d\n", mdio_phy_addr);
	if (mdio_phy_addr == 32)
		return -ENODEV;

	/* Get manufacturer */
	phyid_high = get_mdio_reg(MII_PHYSID1);
	phyid_low = get_mdio_reg(MII_PHYSID2);
	oui = (phyid_high << 6) | (phyid_low >> 10);

	for (ops = &transceivers[0]; ops->oui; ops++) {
		if (ops->oui == oui)
			break;
	}
	DMDIO(printk(KERN_INFO "transceiver ID1=%x ID2=%x, OUI=%x\n",
		     phyid_high, phyid_low, oui));
	transceiver = ops;

	return 0;
}

static void 
set_duplex(struct net_device* dev, enum duplex new_duplex)
{
	if (current_duplex != new_duplex) {
		current_duplex = new_duplex;
		phy_negotiate(dev);
	}
}

static void
clear_network_leds(unsigned long dummy)
{
	if (led_active && time_after(jiffies, led_next_time)) {
		set_network_leds(NO_NETWORK_ACTIVITY);

		/* Set the earliest time we may set the LED */
		led_next_time = jiffies + NET_FLASH_PAUSE;
		led_active = 0;
	}
}

static void
set_network_leds(int active)
{
	int light_leds = (active == NO_NETWORK_ACTIVITY);

	if (!current_speed) {
		/* LED off, link is down */
		LED_NETWORK_SET(LED_OFF);
	}
	else if (light_leds) {
		if (current_speed == 10) {
			LED_NETWORK_SET(LED_ORANGE);
		} else {
			LED_NETWORK_SET(LED_GREEN);
		}
	}
	else {
		LED_NETWORK_SET(LED_OFF);
	}
}

static void
led_flash(void)
{
  	if (!led_active && time_after(jiffies, led_next_time)) {
		/* light the network leds depending on the current speed. */
		set_network_leds(NETWORK_ACTIVITY);

		/* Set the earliest time we may clear the LED */
		led_next_time = jiffies + NET_FLASH_TIME;
		led_active = 1;
		mod_timer(&clear_led_timer, jiffies + HZ/10);
	}
}

static int
oeth_open(struct net_device *dev)
{
	struct oeth_private *cep = (struct oeth_private *)dev->priv;

	volatile oeth_regs *regs = (oeth_regs *)dev->base_addr;

	struct  sk_buff *skb;
	volatile oeth_bd *rx_bd;
	int i;

	D(printk("OPEN: net_device %p\n", dev));

	rx_bd = cep->rx_bd_base;

	for(i = 0; i < OETH_RXBD_NUM; i++) {

		skb = dev_alloc_skb(MAX_FRAME_SIZE);

		if (skb == NULL)
			rx_bd[i].len_status = (0 << 16) |
				OETH_RX_BD_IRQ;
		else {
			rx_bd[i].len_status = (0 << 16) |
				OETH_RX_BD_EMPTY |
				OETH_RX_BD_IRQ;
			skb_reserve(skb, NET_IP_ALIGN);
			rx_bd[i].addr = (int)skb->data;
		}

		cep->rx_skbuff[i] = skb;
	}
	rx_bd[OETH_RXBD_NUM - 1].len_status |= OETH_RX_BD_WRAP;


	/* Install our interrupt handler.
	 */
	request_irq(IRQ_ETH_0, oeth_interrupt, IRQF_DISABLED,
		    "eth", (void *)dev);

	dump_mdio_regs();

	/* Enable receiver and transmiter 
	 */
	regs->moder |= OETH_MODER_RXEN | OETH_MODER_TXEN;

	// Allow transmit
	netif_start_queue(dev);

	return 0;
}

static int
oeth_close(struct net_device *dev)
{
	struct oeth_private *cep = (struct oeth_private *)dev->priv;
	volatile oeth_regs *regs = (oeth_regs *)dev->base_addr;
	int i;

	D(printk("CLOSE: net_device %p\n", dev));

	/* Free interrupt hadler 
	 */
	free_irq(IRQ_ETH_0, (void *)dev);

	/* Disable receiver and transmitesr 
	 */
	regs->moder &= ~(OETH_MODER_RXEN | OETH_MODER_TXEN);	

	/* Free all allocated rx buffers 
	 */
	for (i = 0; i < OETH_RXBD_NUM; i++) {
		if (cep->rx_skbuff[i] != NULL)
			dev_kfree_skb(cep->rx_skbuff[i]);
		
	}

	/* Free all allocated tx buffers 
	 */
	for (i = 0; i < OETH_TXBD_NUM; i++) {
		if (cep->tx_skbuff[i] != NULL)
			dev_kfree_skb(cep->tx_skbuff[i]);
	}

	return 0;
}

static int
oeth_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct oeth_private *cep = (struct oeth_private *)dev->priv;
	volatile oeth_bd *bdp;
	unsigned long flags;

	local_irq_save(flags);

	/* Fill in a Tx ring entry 
	 */
	bdp = cep->tx_bd_base + cep->tx_next;

	if (cep->tx_full) {

		/* All transmit buffers are full.  Bail out.
		 */
		//printk("%s: tx queue full!.\n", dev->name);
		netif_stop_queue(dev);
		local_irq_restore(flags);
		return 1;
	}

	/* Clear all of the status flags.
	 */
	bdp->len_status &= ~OETH_TX_BD_STATS;

	/* If the frame is short, tell CPM to pad it.
	 */
	if (skb->len <= ETH_ZLEN)
		bdp->len_status |= OETH_TX_BD_PAD;
	else
		bdp->len_status &= ~OETH_TX_BD_PAD;

#ifdef DEBUG
	_print("TX\n");
	oeth_print_packet((unsigned long)skb->data, skb->len);
#endif

	/* Set buffer length and buffer pointer.
	 */
	bdp->len_status = (bdp->len_status & 0x0000ffff) |
		(skb->len << 16);
	bdp->addr = (uint)__pa(skb->data);

	/* Save skb pointer.
	 */
	cep->tx_skbuff[cep->tx_next] = skb;

	// Since we have a writeback L1-cache, make sure we clean the
	// skb data area before the DMA tries to fetch anything from it.
	// 
	// We don't need to invalidate since we might want the CPU to
	// touch the area next time and the DMA does not modify the data.

	dmac_clean_range((unsigned long)skb->data, 
			 (unsigned long)skb->data + skb->len);

	cep->tx_next = (cep->tx_next + 1) & OETH_TXBD_NUM_MASK;
	
	//local_irq_save(flags);

	if (cep->tx_next == cep->tx_last)
		cep->tx_full = 1;

	/* Send it on its way.  Tell controller its ready, interrupt when done,
	 * and to put the CRC on the end.
	 */
	bdp->len_status |= (OETH_TX_BD_READY | OETH_TX_BD_IRQ |
			    OETH_TX_BD_CRC);
	
	dev->trans_start = jiffies;

	// Flash with the network LED
	led_flash();

	local_irq_restore(flags);

	return 0;
}

/* The interrupt handler.
 */
static irqreturn_t
oeth_interrupt(int irq, void *dev_id)
{
	struct	net_device *dev = dev_id;
	volatile struct	oeth_private *cep;
	uint	int_events;
	int serviced;
	
	serviced = 0;
	
	cep = (struct oeth_private *)dev->priv;

	/* Get the interrupt events that caused us to be here.
	 */
	int_events = cep->regs->int_src;
	cep->regs->int_src = int_events;

	/* Handle receive event in its own function.
	 */
	if (int_events & (OETH_INT_RXF | OETH_INT_RXE)) {
		D(serviced |= 0x1); 
		oeth_rx(dev_id);
	}

	/* Handle transmit event in its own function.
	 */
	if (int_events & (OETH_INT_TXB | OETH_INT_TXE)) {
		D(serviced |= 0x2);
		oeth_tx(dev_id);
	}

	/* Check for receive busy, i.e. packets coming but no place to
	 * put them. 
	 */
	if (int_events & OETH_INT_BUSY) {
		D(serviced |= 0x4);
		D(printk("b"));
		if (!(int_events & (OETH_INT_RXF | OETH_INT_RXE))) {
			oeth_rx(dev_id);
		}else {
			// Book this as a receiver ring overrun
			cep->stats.rx_over_errors++;
		}
	}

#ifdef DEBUG
	if (serviced == 0)
		printk("\nnot serviced: events 0x%x, irq %d, dev %p\n",
		       int_events, irq, dev_id);
	else 
		printk(" | serviced 0x%x, events 0x%x, irq %d, dev %p\n",
		       serviced, int_events, irq, dev_id);
#endif

	// We handled this interrupt
	return IRQ_RETVAL(1);
}


static void
oeth_tx(struct net_device *dev)
{
	struct	oeth_private *cep;
	volatile oeth_bd *bdp;
	struct	sk_buff *skb;

	D(printk("T"));

	cep = (struct oeth_private *)dev->priv;

	for (;; cep->tx_last = (cep->tx_last + 1) & OETH_TXBD_NUM_MASK) {

		bdp = cep->tx_bd_base + cep->tx_last;

		if ((bdp->len_status & OETH_TX_BD_READY) || 
			((cep->tx_last == cep->tx_next) && !cep->tx_full))
			break;

		/* Check status for errors
		 */
		if (bdp->len_status & OETH_TX_BD_LATECOL) 
			cep->stats.tx_window_errors++;
		if (bdp->len_status & OETH_TX_BD_RETLIM)
			cep->stats.tx_aborted_errors++;
		if (bdp->len_status & OETH_TX_BD_UNDERRUN)
			cep->stats.tx_fifo_errors++;
		if (bdp->len_status & OETH_TX_BD_CARRIER)
			cep->stats.tx_carrier_errors++;
		// Book any tx error into tx_errors
		if (bdp->len_status & (OETH_TX_BD_LATECOL | 
				       OETH_TX_BD_RETLIM |
				       OETH_TX_BD_CARRIER |
				       OETH_TX_BD_UNDERRUN))
			cep->stats.tx_errors++;

		cep->stats.tx_packets++;
		// TODO: maybe increase this only for successful send.
		cep->stats.tx_bytes += bdp->len_status >> 16;
		cep->stats.collisions += (bdp->len_status >> 4) & 0x000f;

		D(printk("TX len_status=0x%08x\n", bdp->len_status));

		skb = cep->tx_skbuff[cep->tx_last];

		/* Free the sk buffer associated with this last transmit.
		*/
		dev_kfree_skb_irq(skb);

		if (cep->tx_full) {
			cep->tx_full = 0;
			// If we had stopped the kernel previously, we
			// can now start sending again since we have
			// at least one buffer free
			if (netif_queue_stopped(dev)) {
				netif_wake_queue(dev);
			}
		}
	}
}

static void
oeth_rx(struct net_device *dev)
{
	struct	oeth_private *cep;
	volatile oeth_bd *bdp;
	struct	sk_buff *skb;
	int	pkt_len;
	int	bad = 0;
	struct	sk_buff *small_skb;

	D(printk("r"));

	cep = (struct oeth_private *)dev->priv;

	/* First, grab all of the stats for the incoming packet.
	 * These get messed up if we get called due to a busy condition.
	 */
	for (;;cep->rx_cur = (cep->rx_cur + 1) & OETH_RXBD_NUM_MASK) {
		bdp = cep->rx_bd_base + cep->rx_cur;

		skb = cep->rx_skbuff[cep->rx_cur];

		// This should not happen, I think.
		if (skb == NULL) {
			skb = dev_alloc_skb(MAX_FRAME_SIZE);
			
			if (skb != NULL) {
				skb_reserve(skb, NET_IP_ALIGN);
				bdp->addr = (unsigned long) skb->data;
				bdp->len_status |= OETH_RX_BD_EMPTY;
			}
			
			continue;
		}
			
		if (bdp->len_status & OETH_RX_BD_EMPTY)
			break;

		D(printk("RX len status: 0x%08x\n", bdp->len_status));
 			
		/* Check status for errors.
		 */
		if (bdp->len_status & (OETH_RX_BD_TOOLONG |
				       OETH_RX_BD_SHORT)) {
			cep->stats.rx_length_errors++;
			bad = 1;
		}
		if (bdp->len_status & OETH_RX_BD_DRIBBLE) {
			cep->stats.rx_frame_errors++;
			bad = 1;
		}
		if (bdp->len_status & OETH_RX_BD_CRCERR) {
			printk("RX crcerr, len_status=0x%08x\n",
			       bdp->len_status);
			cep->stats.rx_crc_errors++;
			bad = 1;
		}
		if (bdp->len_status & OETH_RX_BD_OVERRUN) {
			printk("RX overrun, len_status=0x%08x\n",
			       bdp->len_status);
			cep->stats.rx_fifo_errors++;
			bad = 1;
		}
#if 0
		// 1 if received due to promiscous mode but would
		// otherwise miss the address recognition.
		// not much we should do about that.
		if (bdp->len_status & OETH_RX_BD_MISS) {

		}
#endif
#if 0
		// Davicom...
		if ((bdp->len_status & OETH_RX_BD_LATECOL) && 
		    full_duplex) {
			cep->stats.rx_frame_errors++;
			D(printk("*********DICK**********\n"));
			bad = 1;
		}
#endif
		if (bad) {
			bdp->len_status &= ~OETH_RX_BD_STATS;
			bdp->len_status |= OETH_RX_BD_EMPTY;
			cep->stats.rx_errors++;			
			continue;
		}

		/* Process the incoming frame.
		 */
		pkt_len = bdp->len_status >> 16;

		// Invalidate the buffer since the DMA has touched
		// it and we might have some of it in the L1-cache.
		// No need to clean it.
		dmac_inv_range((unsigned long)__va(bdp->addr), 
			       (unsigned long)__va(bdp->addr) + pkt_len);
		
		// For small packets, copy to a new to save memory.
		// For bigger packets relay the existing copy to the
		// kernel and allocate a new one for the DMA-ring.

		if (pkt_len < RX_COPYBREAK) {
			// Add NET_IP_ALIGN since we do the skb_reserve below
			small_skb = dev_alloc_skb(pkt_len + NET_IP_ALIGN);
			
			if (small_skb) {
				// TODO: this makes the memcpy unaligned.
				// perhaps simply copy aligned but with 
				// an extra 2 bytes at the end, so we include
				// the 2 extra bytes at the head ?
				skb_reserve(small_skb, NET_IP_ALIGN);
				small_skb->dev = dev;
#ifdef DEBUG
				oeth_print_packet(__va(bdp->addr),
						  pkt_len);
#endif
				memcpy(skb_put(small_skb, pkt_len), 
				       (unsigned char *)__va(bdp->addr),
				       pkt_len);
				small_skb->protocol =
					eth_type_trans(small_skb,dev);
				netif_rx(small_skb);
				cep->stats.rx_packets++;
				cep->stats.rx_bytes += pkt_len;
			} else {
				printk("%s: Memory squeeze, dropping packet.\n",
				       dev->name);
	                        cep->stats.rx_dropped++;
			}
			bdp->len_status &= ~OETH_RX_BD_STATS;
			bdp->len_status |= OETH_RX_BD_EMPTY;
		} else {
        		skb->dev = dev;
			skb_put(skb, pkt_len);
			skb->protocol = eth_type_trans(skb, dev);
			netif_rx(skb);

			cep->stats.rx_packets++;
			cep->stats.rx_bytes += pkt_len;

#ifdef DEBUG
			_print("RX long\n");
                        oeth_print_packet(__va(bdp->addr), pkt_len);
#endif
			skb = dev_alloc_skb(MAX_FRAME_SIZE);

			bdp->len_status &= ~OETH_RX_BD_STATS;
        
			if (skb) {
				skb_reserve(skb, NET_IP_ALIGN);
				cep->rx_skbuff[cep->rx_cur] = skb;
				bdp->addr = (int)skb->data;
				bdp->len_status |= OETH_RX_BD_EMPTY;
			} else {
				cep->rx_skbuff[cep->rx_cur] = NULL;	
			}
		}
	}
}

static int calc_crc(char *mac_addr)
{
	int result = 0;
	return (result & 0x3f);
}

static struct net_device_stats *oeth_get_stats(struct net_device *dev)
{
        struct oeth_private *cep = (struct oeth_private *)dev->priv;
 
        return &cep->stats;
}

static void oeth_set_multicast_list(struct net_device *dev)
{
	struct	oeth_private *cep;
	struct	dev_mc_list *dmi;
	volatile oeth_regs *regs;
	int	i;

	unsigned int hash0_temp,hash1_temp;

	cep = (struct oeth_private *)dev->priv;

	/* Get pointer of controller registers.
	 */
	regs = (oeth_regs *)dev->base_addr;

	if (dev->flags & IFF_PROMISC) {
	  
		/* Log any net taps. 
		 */
		printk("%s: Promiscuous mode enabled.\n", dev->name);
		regs->moder |= OETH_MODER_PRO;
	} else {

		regs->moder &= ~OETH_MODER_PRO;

		if (dev->flags & IFF_ALLMULTI) {

			/* Catch all multicast addresses, so set the
			 * filter to all 1's.
			 */
			regs->hash_addr0 = 0xffffffff;
			regs->hash_addr1 = 0xffffffff;
		}
		else if (dev->mc_count) {

			regs->moder |= OETH_MODER_IAM;
//			printk("IAM MODER: 0x%08x\n", regs->moder);
			
                        /* Clear filter and add the addresses in the list.
			 */
			hash0_temp=0x0;
			hash1_temp=0x0;

			dmi = dev->mc_list;

			for (i = 0; i < dev->mc_count; i++) {
				
				unsigned int hash_b;

				/* Only support group multicast for now.
				 */
				if (!(dmi->dmi_addr[0] & 1))
					continue;

				hash_b = (ether_crc(ETH_ALEN, dmi->dmi_addr) & 
					  0xFC000000)>>26;
				if(hash_b >= 32)
					hash1_temp |= 1 << (hash_b - 32 );
				else
					hash0_temp |= 1 << (hash_b );

				dmi = dmi->next;
			}
			regs->hash_addr0 = hash0_temp;
			regs->hash_addr1 = hash1_temp;
		}
	}
}

static int
oeth_set_mac_add(struct net_device *dev, void *p)
{
	struct sockaddr *addr=p;
	volatile oeth_regs *regs;
	int i;

	memcpy(dev->dev_addr, addr->sa_data, dev->addr_len);

	regs = (oeth_regs *)dev->base_addr;

/*
 * old version: sa_data is of type char, will be expanded to int when ored
 * so negative values e.g 0xB5 will be expanded to 0xffffffB5 and ored
 * dev_addr is of type unsigned char and will be expanded to unsigned int 
 * which is ok.
 *
 */
	regs->mac_addr1 =
		dev->dev_addr[0] << 8 	|
		dev->dev_addr[1];

	regs->mac_addr0 = 
		dev->dev_addr[2] << 24 	|
		dev->dev_addr[3] << 16 	|
		dev->dev_addr[4] << 8 	|
		dev->dev_addr[5];

	// show it in the log as well

	printk(KERN_INFO "%s: changed MAC to ", dev->name);
	for (i = 0; i < 5; i++)
		printk("%02X:", dev->dev_addr[i]);
	printk("%02X\n", dev->dev_addr[i]);

	return 0;
}


/* Initialize the Open Ethernet MAC.
 */

int __init 
oeth_probe(void)
{
	struct oeth_private *cep;
	volatile oeth_regs *regs;
	volatile oeth_bd *tx_bd, *rx_bd;
	int i, j, k;
#ifdef SRAM_BUFF
	unsigned long mem_addr = SRAM_BUFF_BASE;
#else
	unsigned long mem_addr;
#endif
	struct net_device *dev;
	int err;

//	printk(KERN_INFO "%s\n", __DRIVER_NAME);

	dev = alloc_etherdev(sizeof(struct oeth_private));

	if (!dev)
		return -ENOMEM;

	if (dev->base_addr)
		return -ENODEV;

	cep = (struct oeth_private *)dev->priv;

	/* Initialize the device structure. 
	 */
	if (dev->priv == NULL) {
		cep = (struct oeth_private *)kmalloc(sizeof(*cep), GFP_KERNEL);
		dev->priv = cep;
		if (dev->priv == NULL)
			return -ENOMEM;
	}

	__clear_user(cep,sizeof(*cep));

	/* Get pointer ethernet controller configuration registers.
	 */
	cep->regs = (oeth_regs *)(OETH_REG_BASE);
	regs = (oeth_regs *)(OETH_REG_BASE);

	// Start MAC and PHY
	ETH_CLOCK_START;

	/* Reset the controller.
	 */
	regs->moder = OETH_MODER_RST;	/* Reset ON */
	regs->moder &= ~OETH_MODER_RST;	/* Reset OFF */

	/* Setting TXBD base to OETH_TXBD_NUM.
	 */
	regs->tx_bd_num = OETH_TXBD_NUM;
	
	/* Initialize TXBD pointer
	 */
	cep->tx_bd_base = (oeth_bd *)OETH_BD_BASE;
	tx_bd = (volatile oeth_bd *)OETH_BD_BASE;

	/* Initialize RXBD pointer
	 */
	cep->rx_bd_base = ((oeth_bd *)OETH_BD_BASE) + OETH_TXBD_NUM;
	rx_bd = ((volatile oeth_bd *)OETH_BD_BASE) + OETH_TXBD_NUM;

	/* Initialize transmit pointers.
	 */
	cep->rx_cur = 0;
	cep->tx_next = 0;
	cep->tx_last = 0;
	cep->tx_full = 0;

	/* Set min/max packet length 
	 */
	regs->packet_len = 0x00400600;

	/* Set IPGT register to recomended value 
	 */
	regs->ipgt = 0x00000012;

	/* Set IPGR1 register to recomended value 
	 */
	regs->ipgr1 = 0x0000000c;

	/* Set IPGR2 register to recomended value 
	 */
	regs->ipgr2 = 0x00000012;

	/* Set COLLCONF register to recomended value 
	 */
	regs->collconf = 0x000f003f;

	/* Set control module mode 
	 */
	// Enable PAUSE-frame flow-control
	regs->ctrlmoder = OETH_CTRLMODER_TXFLOW | OETH_CTRLMODER_RXFLOW;
	//regs->ctrlmoder = OETH_CTRLMODER_PASSALL;

  /* Set PHY to show Tx status, Rx status and Link status */
/*
//  regs->miiaddress = 20<<8;
  regs->miiaddress = (20<<8) | 16;
  regs->miitx_data = 0x1422;
  regs->miicommand = OETH_MIICOMMAND_WCTRLDATA;
*/
/*
  // Set PHY in loopback mode
  printk("Setting PHY in loopback mode\n");
  regs->miiaddress = (0<<8) | 16;
  regs->miitx_data = (1<<14) | (1<<8);
  regs->miicommand = OETH_MIICOMMAND_WCTRLDATA;
  */ 

 	/* Initialize TXBDs.
	 */
	for(i = 0; i < OETH_TXBD_NUM; i++) {

		cep->tx_skbuff[i] = NULL;

		tx_bd[i].len_status = (0 << 16) | OETH_TX_BD_PAD |
			OETH_TX_BD_CRC | OETH_RX_BD_IRQ;
		tx_bd[i].addr = 0;
	}
	tx_bd[OETH_TXBD_NUM - 1].len_status |= OETH_TX_BD_WRAP;

	/* Initialize RXBDs.
	 */
	for(i = 0; i < OETH_RXBD_NUM; i++) {


		rx_bd[i].len_status = (0 << 16) | OETH_RX_BD_IRQ;

		cep->rx_skbuff[i] = NULL;

		rx_bd[i].addr = 0;
	}
	rx_bd[OETH_RXBD_NUM - 1].len_status |= OETH_RX_BD_WRAP;

	/* Set default ethernet station address.
	 */
	dev->dev_addr[0] = MACADDR0;
	dev->dev_addr[1] = MACADDR1;
	dev->dev_addr[2] = MACADDR2;
	dev->dev_addr[3] = MACADDR3;
	dev->dev_addr[4] = MACADDR4;
	dev->dev_addr[5] = MACADDR5;

	regs->mac_addr1 = MACADDR0 << 8 | MACADDR1;
	regs->mac_addr0 = 
		MACADDR2 << 24 | MACADDR3 << 16 |
		MACADDR4 << 8  | MACADDR5;
	
	/* Clear all pending interrupts 
	 */
	regs->int_src = 0xffffffff;

	/* Promisc, IFG, CRCEn
	 */
//	regs->moder |= OETH_MODER_PAD | OETH_MODER_IFG | OETH_MODER_CRCEN;
	// start in full duplex mode per default (see full_duplex var below)
	regs->moder |= OETH_MODER_PAD | OETH_MODER_IFG | OETH_MODER_CRCEN
		| OETH_MODER_FULLD;
//	regs->moder |= OETH_MODER_PAD | OETH_MODER_IFG | OETH_MODER_CRCEN | OETH_MODER_FULLD | OETH_MODER_PRO;
//	regs->moder |= OETH_MODER_PAD | OETH_MODER_IFG | OETH_MODER_CRCEN | OETH_MODER_FULLD | OETH_MODER_PRO | OETH_MODER_LOOPBCK;
//	regs->moder |= OETH_MODER_RECSMALL | OETH_MODER_PAD | OETH_MODER_IFG | OETH_MODER_CRCEN | OETH_MODER_FULLD | OETH_MODER_PRO | OETH_MODER_LOOPBCK;
//  printk("Init MODER: 0x%08x\n", regs->moder);

	/* Enable interrupt sources.
	 */
	regs->int_mask = OETH_INT_MASK_TXB 	| 
		OETH_INT_MASK_TXE 	| 
		OETH_INT_MASK_RXF 	| 
		OETH_INT_MASK_RXE 	|
		OETH_INT_MASK_BUSY 	|
		OETH_INT_MASK_TXC	|
		OETH_INT_MASK_RXC;

	/* Fill in the fields of the device structure with ethernet values. 
	 */
//	ether_setup(dev);

	dev->base_addr = (unsigned long)OETH_REG_BASE;

	/* The Open Ethernet specific entries in the device structure. 
	 */
	dev->open = oeth_open;
	dev->hard_start_xmit = oeth_start_xmit;
	dev->stop = oeth_close;
	dev->get_stats = oeth_get_stats;
	dev->set_multicast_list = oeth_set_multicast_list;
	dev->set_mac_address = oeth_set_mac_add;

	// Probe PHY
	if (probe_transceiver(dev)) {
		free_netdev(dev);
		return -EAGAIN;
	}

	phy_reset();

	/* Register device */
	err = register_netdev(dev);
	if (err) {
		free_netdev(dev);
		return err;
	}

	// Initialize speed indicator stuff

	led_next_time = jiffies;
	current_speed = 10;
#ifdef CONFIG_ARGUS_FPGA
	current_speed_selection = 10;
#else
	current_speed_selection = 0; // auto 10/100
#endif
	speed_timer.expires = jiffies + NET_LINK_UP_CHECK_INTERVAL;
	speed_timer.function = check_speed;
	speed_timer.data = (unsigned long)dev;
	add_timer(&speed_timer);

	clear_led_timer.function = clear_network_leds;

	full_duplex = 1;
	current_duplex = full;
	duplex_timer.expires = jiffies + NET_DUPLEX_CHECK_INTERVAL;		
	duplex_timer.function = check_duplex;
	duplex_timer.data = (unsigned long)dev;
	add_timer(&duplex_timer);

	// Start autonegotiation
	phy_negotiate(dev);

	ETH_CLOCK_STOP;

	printk("%s: Open Ethernet Core Version 1.0\n", dev->name);

	return 0;
}

module_init(oeth_probe);
