// SSI driver for Argus2/3
//
// Copyright (c) 2004-2006 Axis Communications AB
//
// Authors:   Bjorn Wesen

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/timer.h>

#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <asm/pgalloc.h>
#include <asm/arch/argus.h>
#include <asm/arch/argdma.h>
#include <asm/arch/dma_alloc.h>
#include <asm/arch/gpio.h>
#include <asm/arch/ssi.h>
#include <asm/arch/sys_time.h>

#define DBG(x)
#define DBG2(x)

#define USE_CLK_GATING 0

#ifdef CONFIG_ARGUS_3
#define BASE_CLOCK 75000000 /* HCLK is 150/2 MHz */
#else
#define BASE_CLOCK 97500000 /* HCLK is 195/2 MHz */
#endif

#define SYNC_SERIAL_MAJOR 125

// Number of software buffers and their sizes

#define NUM_IN_BUFFERS 4
#define NUM_OUT_BUFFERS 4

// Number of hardware DMA slots
#define NUM_DMA_SLOTS 2

// The various states a buffer can be in and if its relevant for TX or RX
enum bufstate {
	BUF_IDLE,        // not used
	BUF_RUNNING,     // issued to the DMA (rx or tx)
	BUF_DONE,        // DMA is done, has valid data (rx)
	BUF_COPYING,     // usermode is copying the contents (rx or tx)
	BUF_LOADED       // TX buffer loaded with data waiting for DMA (tx)
};

struct ssi_buffer {
	char *buffer;

	volatile enum bufstate state;
	struct timeval timestamp;  // set when grabbed, for input buffers
};

struct ssi_port {
	struct ssi_buffer inbuf[NUM_IN_BUFFERS];
	struct ssi_buffer outbuf[NUM_OUT_BUFFERS];

	volatile int cur_in;  // next buffer to be started by the DMA
	volatile int cur_out;

	int dma_in_ch, dma_out_ch;     // Allocated DMA channel numbers
	volatile int dma_in_running, dma_out_running; // Set if DMA is running

	wait_queue_head_t wq_in, wq_out;

	int opened; // Enforce one user limit
	int input_mode; // halting, continuous
	// set to use only one input buffer
	// (in reality, all buffers are queued but when usermode starts
	//  to read, the newest is read and the older are flushed)
	int one_buffer_input_mode;
};

static int ssi_start_dma_in(struct ssi_port *p);
static int ssi_start_dma_out(struct ssi_port *p);

// Get a timeval that is not affected by date-sets, which we can use in timestamps

extern int argus_gettimeoffset(void);  // included from asm/arch/time.h

static void
ssi_get_monotonic_time(struct timeval *tv)
{
#ifdef CLOCK_MONOTONIC
	struct timespec tp;

	do_posix_clock_monotonic_gettime(&tp);
	tv->tv_sec = tp.tv_sec;
	tv->tv_usec = tp.tv_nsec / 1000;
#else
	static const struct timeval wrap_time = {
		.tv_sec = (ULONG_MAX+1ULL) / HZ,
		.tv_usec = ((ULONG_MAX+1ULL) % HZ) * (1000000 / HZ)
	};
	static struct timeval total_wrap_time;
	static unsigned long last;
	int flags;

	local_irq_save(flags);

	tv->tv_sec = jiffies / HZ;
	tv->tv_usec = argus_gettimeoffset() + (jiffies % HZ) * (1000000 / HZ);

	if (tv->tv_usec > 1000000) {
		tv->tv_sec++;
		tv->tv_usec -= 1000000;
	}

	/* Check if jiffies has wrapped. */
	if (jiffies < last) {
		timeradd(&total_wrap_time, &wrap_time, &total_wrap_time);
	}
	last = jiffies;
	local_irq_restore(flags);

	timeradd(tv, &total_wrap_time, tv);
#endif
}

// Enable or disable the SSI transmitter
//
// When the transmitter is disabled, the SSI-block will start to loopback
// data from SSI_RX to SSI_TX, which we don't want. So when we disable
// the transmitter we switch over the SSI_TX pin from ALT_FUNC to GPIO_MODE_OUT
// and back again when the transmitter is enabled.

static void
enable_transmitter(int enable)
{
	if(enable) {
		// Enable the SSI block clock and the transmitter
		SOC_CLK.ENABLE = CLK_SSI;
		SOC_SSI.SSIB |= SET_SSIB_TXEN(1);
		gpio_set_mode(GPIO_SSI_TX, GPIO_MODE_ALT_FUNC, 0);
	} else {
		// Disable the transmitter
		SOC_SSI.SSIB &= ~SET_SSIB_TXEN(1);
		// and force the SSI_TX output to be zero.
		gpio_set_mode(GPIO_SSI_TX, GPIO_MODE_OUT, 0);
		
#if USE_CLK_GATING
		// Turn off the SSI-clock to save power
		// TODO: check if input is still running
		SOC_CLK.DISABLE = CLK_SSI;
#endif
	}
}

// Allocate and initialize a buffer

static int
setup_ssi_buffer(struct ssi_buffer *buf)
{
	if((buf->buffer = kmalloc(SSI_BUFFER_SIZE, GFP_KERNEL)) == 0) {
		printk("cant alloc buffer\n");
		return -ENOMEM;
	}
	buf->state = BUF_IDLE;
	return 0;
}

// Free a buffer

static void
free_ssi_buffer(struct ssi_buffer *buf)
{
	kfree(buf->buffer);
}

// A global instance of our single port

static struct ssi_port port0;

// Find the oldest non-running buffer with active
// data (in the DONE state). If we are in one-buffer mode, purge
// all DONE buffers except the latest and return that.

static int
get_rx_buf(struct ssi_port *p)
{
	int bidx;

	bidx = p->cur_in;  // Oldest buffer that could possibly contain data

	while(p->inbuf[bidx].state != BUF_DONE) {
 		int next = (bidx + 1) % NUM_IN_BUFFERS;
		if(next == p->cur_in) {
			// Looped around, nothing found.
			bidx = -1;
			break;
		}
		bidx = next;
	}

	// If we have a DONE buffer now, and we're in one-buffer mode, mark it
	// and all subsequent DONE buffers as IDLE except the last (latest) one.

	if(bidx >= 0 && p->one_buffer_input_mode) {
		int next = (bidx + 1) % NUM_IN_BUFFERS;
		// We got a DONE buffer - purge it and try again 
		while(next != p->cur_in && p->inbuf[next].state == BUF_DONE) {
			p->inbuf[bidx].state = BUF_IDLE;
			bidx = next;
			next = (bidx + 1) % NUM_IN_BUFFERS; // it hurts to write code like this..
		}
	}

	return bidx;
}

// Find the next non-running buffer which is empty (IDLE)

static int
get_tx_buf(struct ssi_port *p)
{
	int bidx;

	bidx = p->cur_out;  // Oldest buffer that could be idle

	while(p->outbuf[bidx].state != BUF_IDLE) {
 		int next = (bidx + 1) % NUM_OUT_BUFFERS;
		if(next == p->cur_out) {
			// Looped around, nothing found.
			bidx = -1;
			break;
		}
		bidx = next;
	}

	return bidx;
}

// Get an rx buffer full of data, block if it is not available yet.
//
// If the port input DMA is not running, it is started and then we
// wait until the first buffer is ready.
//
// If we cannot keep up with the DMA, the DMA will shut off and
// purge all non-read buffers, to avoid reading non-sequential
// data upon the next read (unless the input is set to continuous mode
// when the purge is not done until we start reading again after a lag,
// save one buffer)
//
// The timestamp of the copied buffer is returned in timestamp

static ssize_t
ssi_get_rx(struct ssi_port *p, char *buf, struct timeval *timestamp)
{
	unsigned long flags;
	int bidx, error;
	struct ssi_buffer *b;

 retry:		
	local_irq_save(flags);

	bidx = get_rx_buf(p);
	
	// If we have no data, block and try again.
		
	if(bidx < 0) {
		// Do we need to start the DMA ?
		
		if(!p->dma_in_running) {
			DBG(printk("SSI: starting RX DMA\n"));
			// Enable SSI-block clock and the receiver
			SOC_CLK.ENABLE = CLK_SSI;
			SOC_SSI.SSIB |= SET_SSIB_RXEN(1);
			ssi_start_dma_in(p);
		}
		
		// The interrupts need to be disabled up until after this
		// sleep, to avoid the race where the irq routine wakes
		// the queue before we have time to sleep.
		
		interruptible_sleep_on(&p->wq_in);
		
		local_irq_restore(flags);

		// Might not always get data here, could be a signal
		// as well, then we bail out immediately.
		
		if(signal_pending(current)) {
			DBG(printk("ssi: get_rx bailing due to signal\n"));
			return -EINTR;
		}
		
		// Go back and get the new data.
		goto retry;
	}

	// We can leave the one-buffer inputmode now since we got some data
	// (i.e. from now on subsequent reads will read the oldest data)

	p->one_buffer_input_mode = 0;
		
	// Lock it, restore interrupts and copy the data.

	b = &p->inbuf[bidx];

	b->state = BUF_COPYING;

	local_irq_restore(flags);
	
	error = copy_to_user(buf, b->buffer, SSI_BUFFER_SIZE);

	// Done with the buffer, mark it idle so the DMA can grab it again
	// The DMA callback can have grabbed it while we were copying, in
	// that case leave the state and mark the call as unsuccessful

	if(b->state != BUF_COPYING)
		return -EAGAIN;

	b->state = BUF_IDLE;

	if(error) {
		DBG(printk("ssi: rx user buffer faulted\n"));
		return -EFAULT;
	}
	
	// Copy out the timestamp of the buffer.

	if(timestamp)
		*timestamp = b->timestamp;

	return SSI_BUFFER_SIZE;
}

// Add the buffer to the tx queue. If it is full, we block until another
// data buffer fits.
//
// If the output port DMA is not running, it is started and then we
// return directly while it is outputting.
//
// If we cannot keep up with the DMA, the DMA will shut off temporarily.

static ssize_t
ssi_put_tx(struct ssi_port *p, char *buf)
{
	unsigned long flags;
	int bidx, error;
	struct ssi_buffer *b;

 retry:	
	local_irq_save(flags);
	
	bidx = get_tx_buf(p);
	
	// If we have no free buffers, block and wait.
		
	if(bidx < 0) {
		// The interrupts need to be disabled up until after this
		// sleep, to avoid the race where the irq routine wakes
		// the queue before we have time to sleep.

		interruptible_sleep_on(&p->wq_out);
		
		local_irq_restore(flags);
		
		// Might not always get buffers here, could be a signal
		// as well, then we bail out immediately.
		
		if(signal_pending(current)) {
			DBG(printk("ssi: put_tx bailing due to signal\n"));
			return -EINTR;
		}
		
		// Go back and look for the new buffer
		goto retry;
	}
	
	// Lock it, restore interrupts and copy the data.

	b = &p->outbuf[bidx];

	b->state = BUF_COPYING;

	local_irq_restore(flags);

	error = copy_from_user(b->buffer, buf, SSI_BUFFER_SIZE);

	b->state = BUF_LOADED;

	if(error) {
		DBG(printk("ssi: tx user buffer faulted\n"));
		b->state = BUF_IDLE;
		return -EFAULT;
	}

	// Now we have at least one valid buffer in the TX queue. Check
	// if the DMA is running otherwise we have to start it.

	if(!p->dma_out_running) {
		DBG(printk("SSI: starting TX DMA\n"));
		enable_transmitter(1);
		ssi_start_dma_out(p);
	}

	return SSI_BUFFER_SIZE;
}

static int
ssi_rx(struct ssi_port *p, unsigned long arg)
{
	struct ssi_request req;
	int read_bytes, ret;

	// Copy the request structure from user-mode
	
	ret = copy_from_user(&req,
			     (struct ssi_request *)arg,
                             sizeof(struct ssi_request));
        
        if(ret)
                return -EFAULT;
	
	// Do the actual read

	read_bytes = ssi_get_rx(p, req.buf, &req.timestamp);

	if(read_bytes < 0)
		return read_bytes;

	// Copy back the request

	ret = copy_to_user((struct ssi_request *) arg,
			   &req,
			   sizeof(struct ssi_request)); 

	if(ret)
		return -EFAULT;

	// Return the number of bytes read

	return read_bytes;
}

static int
ssi_tx(struct ssi_port *p, unsigned long arg)
{
	struct ssi_request req;
	int written_bytes, ret;

	// Copy the request structure from user-mode
	
	ret = copy_from_user(&req,
			     (struct ssi_request *)arg,
                             sizeof(struct ssi_request));
        
        if(ret)
                return -EFAULT;
	
	// Do the actual write

	written_bytes = ssi_put_tx(p, req.buf);

	return written_bytes;
}

static int 
ssi_ioctl(struct inode *inode,
	  struct file *file,
	  unsigned int cmd, 
	  unsigned long arg)
{
	struct ssi_port *p = (struct ssi_port *)file->private_data;
	int dsize;

	switch(cmd) {
		case SSI_SET_SSIA:

			if(p->dma_in_running || p->dma_out_running)
				printk("SSI: WARNING! Changing SSIA on running port!\n");

			// SSIA controls clockspeed/dividers and transfer size

			SOC_SSI.SSIA = arg;

			break;

		case SSI_SET_SSIB:

			if(p->dma_in_running || p->dma_out_running)
				printk("SSI: WARNING! Changing SSIB on running port!\n");

			// SSIB controls data, clock and sync format

			SOC_SSI.SSIB = arg;

			break;

		case SSI_INPUT:
			// Blocking read (but returns directly if there is anything
			// in the queue)
			return ssi_rx(p, arg);

		case SSI_OUTPUT:
			// Blocking write (but returns directly if there is an empty
			// buffer in the queue to fill)
			return ssi_tx(p, arg);

		case SSI_SET_DMA_WIDTH:

			if(p->dma_in_running || p->dma_out_running)
				printk("SSI: WARNING! Changing DMA width on running port!\n");

			if(arg == 1)
				dsize = DMA_SIZE_BYTE;
			else if(arg == 2)
				dsize = DMA_SIZE_SHORT;
			else
				dsize = DMA_SIZE_WORD;

			dma_setup(p->dma_in_ch,
				  DMA_PORT_SSI_OUT, 1, 1, 
				  dsize, DMA_DEV2MEM);

			dma_setup(p->dma_out_ch,
				  DMA_PORT_SSI_IN, 1, 1, 
				  dsize, DMA_MEM2DEV);
			break;

		case SSI_SET_INPUT_MODE:
			p->input_mode = arg;

			if(arg == SSI_INPUT_MODE_HALT) {
				p->one_buffer_input_mode = 0;
			} else if(arg == SSI_INPUT_MODE_CONTINUOUS) {
				// start the receiver if it is not
				// running
				if(!p->dma_in_running) {
					DBG(printk("SSI: starting RX DMA\n"));
					// Enable SSI-block clock and the receiver              
					SOC_CLK.ENABLE = CLK_SSI;
					SOC_SSI.SSIB |= SET_SSIB_RXEN(1);
					ssi_start_dma_in(p);
				}
			}
			break;
	}	

	return 0;
}

// Should be called with irq's disabled.
//
// Returns the number of new issued buffers, if this is 0
// and we call from a DMA done context, it means that the
// we have caught up with the users datareading.

static int
ssi_start_dma_in(struct ssi_port *p)
{
	int started = 0;

	// Enable the input DMA channel

	dma_enable(p->dma_in_ch, 1, 1);
	dma_irq_enable(p->dma_in_ch, 1);
	p->dma_in_running = 1;

	// Start the input DMA on the buffer pointed
	// to by the cur_in index. Issue as many unused buffers
	// as will fit in the DMA queue.
	//
	// If we're in the continous mode, make sure we start at
	// least one new buffer, even if it means that we have to
	// write over the oldest. A possible complication can arise
	// in that scenario - we might restart a buffer that has taken
	// too long to copy, i.e. it was in the BUF_COPYING state. We
	// can't avoid that, the buffer will be garbishly copied. The
	// copying routine will check if it was restarted under its feet
	// and leave it at BUF_RUNNING.

	while(((p->inbuf[p->cur_in].state == BUF_IDLE) ||
	       (p->input_mode == SSI_INPUT_MODE_CONTINUOUS && 
		started < 1 &&
		p->inbuf[p->cur_in].state != BUF_RUNNING)) &&
	      dma_buffer_status(p->dma_in_ch, 0) < NUM_DMA_SLOTS) {

		DBG2(printk("adding buf at 0x%p\n", p->inbuf[p->cur_in].buffer));

		ct200_dma_inv_range((int)p->inbuf[p->cur_in].buffer, 
				    (int)p->inbuf[p->cur_in].buffer + SSI_BUFFER_SIZE);

		dma_add_buffer(p->dma_in_ch,
			       virt_to_phys(p->inbuf[p->cur_in].buffer),
			       SSI_BUFFER_SIZE,
			       DMA_FLAGS_CACHEABLE | DMA_FLAGS_BUFFERABLE);
		started++;

		// If we are lagging and restarted an old buffer, go into the one-buffer
		// mode 
		
		if(p->inbuf[p->cur_in].state != BUF_IDLE)
			p->one_buffer_input_mode = 1;

		p->inbuf[p->cur_in].state = BUF_RUNNING;

		p->cur_in = (p->cur_in + 1) % NUM_IN_BUFFERS;
	}

	// Notice that when we leave here, cur_in points to the *next* buffer
	// to eventually be started by the input DMA. Maybe it should be renamed
	// to next_in :) 
	
	return started;
}

// Should be called with irq's disabled.
//
// Returns the number of new issued buffers, if this is 0
// and we call from a DMA done context, it means that the
// user is too slow.

static int
ssi_start_dma_out(struct ssi_port *p)
{
	int started = 0;

	// Enable the output DMA channel

	dma_enable(p->dma_out_ch, 1, 1);
	dma_irq_enable(p->dma_out_ch, 1);
	p->dma_out_running = 1;

	// Start the output DMA on the buffer pointed
	// to by the cur_out index. Issue as many loaded buffers
	// as will fit in the DMA queue.

	while(p->outbuf[p->cur_out].state == BUF_LOADED &&
	      dma_buffer_status(p->dma_out_ch, 0) < NUM_DMA_SLOTS) {

		DBG2(printk("adding outbuf at 0x%p\n", p->outbuf[p->cur_out].buffer));

		dma_add_buffer(p->dma_out_ch,
			       virt_to_phys(p->outbuf[p->cur_out].buffer),
			       SSI_BUFFER_SIZE,
			       DMA_FLAGS_CACHEABLE | DMA_FLAGS_BUFFERABLE);
		started++;

		p->outbuf[p->cur_out].state = BUF_RUNNING;

		p->cur_out = (p->cur_out + 1) % NUM_OUT_BUFFERS;
	}

	// Notice that when we leave here, cur_out points to the *next* buffer
	// to eventually be started by the output DMA. Maybe it should be renamed
	// to next_out :) 
	
	return started;
}

// Called when a buffer is completely read. Mark it done.
// Wake up waiters and issue new buffers.
// If there are no free buffers, halt and
// clear everything since we do not want the user
// to read data with gaps in it.

static void
ssi_in_callback(int ch)
{
	int i, bidx;
	struct ssi_port *p = &port0;

#ifdef CONFIG_ARGUS_3
        // Ack
        SOC_DMA.CHANNEL[ch].CHACTRL = SET_DMA_CHACTRL_IRQ(1);
#endif

	// No need to safetycheck the channel number since we're
	// the only callback for the input DMA channel.

	// Find the buffer that completed - it is the oldest
	// running buffer.

	bidx = p->cur_in;
	while(p->inbuf[bidx].state != BUF_RUNNING) {
 		int next = (bidx + 1) % NUM_IN_BUFFERS;
		if(next == p->cur_in) {
			// Looped around, nothing found.
			bidx = -1;
			break;
		}
		bidx = next;
	}

	if(bidx < 0) {
		int j;
		// Should always have something here.
		printk("SSI: BUG, bogus in completed buffer. cur_in %d\n", p->cur_in);
		for(j = 0; j < NUM_IN_BUFFERS; j++)
			printk("  buf %d state %d\n", j, p->inbuf[j].state);
	}

	// Mark it done and set its timestamp

	p->inbuf[bidx].state = BUF_DONE;
		
	ssi_get_monotonic_time(&p->inbuf[bidx].timestamp);

	// Wake up any waiters

	wake_up_interruptible(&p->wq_in);

	// Issue more buffers to the DMA

	if(!ssi_start_dma_in(p)) {
		// Halt and clear all done flags
		dma_enable(ch, 0, 0);
		dma_irq_enable(ch, 0);
		dma_flush_buffer(ch);
		p->dma_in_running = 0;
		for(i = 0; i < NUM_IN_BUFFERS; i++) {
			// Leave the buffers that are already
			// copying to usermode.
			if(p->inbuf[i].state != BUF_COPYING)
				p->inbuf[i].state = BUF_IDLE;
		}
		// Disable the receiver
		SOC_SSI.SSIB &= ~SET_SSIB_RXEN(1);
#if USE_CLK_GATING
		// Turn off the SSI-clock to save power
		// TODO: check if output is still running
		SOC_CLK.DISABLE = CLK_SSI;
#endif
		DBG(printk("SSI: input stopped and flushed\n"));
	}
}

// Called when a buffer is completely written. Mark it idle.
// Wake up waiters and issue new buffers.
// If there are no more buffers, halt the DMA.

static void
ssi_out_callback(int ch)
{
	int bidx;
	struct ssi_port *p = &port0;

#ifdef CONFIG_ARGUS_3
        // Ack
        SOC_DMA.CHANNEL[ch].CHACTRL = SET_DMA_CHACTRL_IRQ(1);
#endif

	// No need to safetycheck the channel number since we're
	// the only callback for the input DMA channel.

	// Find the buffer that completed - it is the oldest
	// running buffer.

	bidx = p->cur_out;  // Oldest buffer that could be idle
	while(p->outbuf[bidx].state != BUF_RUNNING) {
 		int next = (bidx + 1) % NUM_OUT_BUFFERS;
		if(next == p->cur_out) {
			// Looped around, nothing found.
			bidx = -1;
			break;
		}
		bidx = next;
	}

	if(bidx < 0) // Should always have something here.
		printk("SSI: BUG, bogus out completed out buffer\n");

	// Mark it idle, we're done with this buffer now.

	p->outbuf[bidx].state = BUF_IDLE;
		
	// Wake up any waiters

	wake_up_interruptible(&p->wq_out);

	// Issue more buffers to the DMA

	if(!ssi_start_dma_out(p)) {
		// Halt and clear all done flags
		dma_enable(ch, 0, 0);
		dma_irq_enable(ch, 0);
		// There might be remaining data in the DMA HW queue
		dma_flush_buffer(ch);
		p->dma_out_running = 0;
		// Disable the transmitter
		enable_transmitter(0);
		DBG(printk("SSI: output stopped\n"));
	}
}


// Open the port.
// 
// Allocate buffers and DMA channels
//
// TODO: possibly check open direction and allocate in/out/both depending.
// (there is a finite number of DMA channels in the chip)

static int
ssi_open(struct inode *inode, struct file *filp)
{
	int i;

	if(port0.opened)
		return -EBUSY;

	memset(&port0, 0, sizeof(struct ssi_port));

        init_waitqueue_head(&port0.wq_in);
        init_waitqueue_head(&port0.wq_out);
        
	// Probably only need a small FIFO so request the smallest
	// available.

	if((port0.dma_in_ch = dma_alloc(ssi_in_callback, 0)) < 0) {
		printk("ssi: dma in alloc error");
		return -ENOMEM;
	}

	if((port0.dma_out_ch = dma_alloc(ssi_out_callback, 0)) < 0) {
		printk("ssi: dma out alloc error");
		dma_free(port0.dma_in_ch);
		return -ENOMEM;
	}

	DBG(printk("ssi open: got dma channels in %d out %d\n",
		   port0.dma_in_ch, port0.dma_out_ch));

	// Prepare input buffers

	for(i = 0; i < NUM_IN_BUFFERS; i++) {
		if(setup_ssi_buffer(&port0.inbuf[i]) < 0) {
			dma_free(port0.dma_in_ch);
			dma_free(port0.dma_out_ch);
			return -ENOMEM;
		}
	}

	// Prepare output buffers

	for(i = 0; i < NUM_OUT_BUFFERS; i++) {
		if(setup_ssi_buffer(&port0.outbuf[i]) < 0) {
			dma_free(port0.dma_in_ch);
			dma_free(port0.dma_out_ch);
			// We lose the allocated input buffers. I'm too lazy to
			// free them, the memory is out anyway...
			return -ENOMEM;
		}
	}
  
	// Make sure the DMA channel is disabled but
	// correctly configured. dma_setup will leave the
	// channel in a flushed/reset state.
	//
	// SSI input is read from the SSI_OUT port and vice versa!
	//
	// Default transfer-size is word (16 bits). 
	//
	// Use highest prio, because the transfers are so very short but
	// the SSI block does not have any internal FIFO's so it's very
	// important to actually serve them with low latency or we'll get
	// nasty pops in the audio.

	dma_setup(port0.dma_in_ch,
		  DMA_PORT_SSI_OUT, 1, 3, 
		  DMA_SIZE_SHORT, DMA_DEV2MEM);

	dma_setup(port0.dma_out_ch,
		  DMA_PORT_SSI_IN, 1, 3, 
		  DMA_SIZE_SHORT, DMA_MEM2DEV);
	
	port0.opened = 1;

        filp->private_data = &port0;

	return 0;
}

// Close the port. 

static int 
ssi_release(struct inode *inode, struct file *filp)
{
	int i;
	struct ssi_port *p = (struct ssi_port *)filp->private_data;

	for(i = 0; i < NUM_IN_BUFFERS; i++)
		free_ssi_buffer(&p->inbuf[i]);

	for(i = 0; i < NUM_OUT_BUFFERS; i++)
		free_ssi_buffer(&p->outbuf[i]);

	dma_free(p->dma_in_ch);
	dma_free(p->dma_out_ch);

	p->opened = 0;

	return 0;
}

// poll checks if at least one buffer is ready for reading or writing

static unsigned int 
ssi_poll(struct file *filp,
	 struct poll_table_struct *wait)
{
	struct ssi_port *p = (struct ssi_port *)filp->private_data;
	unsigned int mask;

	// Add both in and out waitqueues to the pollqueues

        poll_wait(filp, &p->wq_in, wait);
        poll_wait(filp, &p->wq_out, wait);

	// Then build a mask of what we the stat is right now and return

	mask = 0;

        if(get_rx_buf(p) >= 0) {
		// There is at least one buffer completed to read
                mask |= POLLIN | POLLRDNORM;
        }
        if(get_tx_buf(p) >= 0) {
		// There is at least one buffer free to write
                mask |= POLLOUT | POLLWRNORM;
        }

        return mask;
}

static struct file_operations ssi_fops = {
	.owner   = THIS_MODULE,
	// Read/write handled only by the ioctl now
	//.write   = ssi_write,
	//.read    = ssi_read,
	.poll    = ssi_poll,
	.ioctl   = ssi_ioctl,
	.open    = ssi_open,
	.release = ssi_release
};

static int __init 
ssi_init(void)
{
	if (register_chrdev(SYNC_SERIAL_MAJOR, "ssi", 
			    &ssi_fops) <0 ) {
		printk("ssi: no major!\n");
		return -EBUSY;
	}

	port0.opened = 0;

#ifndef CONFIG_ARGUS_3
	// Setup the clock generator output pin, CLKGEN1.

	// For 16 khz codec operation, we need a clock
	// which is 250 times that value for AIC23B. 
	//
	// 16000*250=4 mhz
	//
	// With a system clock of 195 mhz in Argus2, we need to
	// divide by 48.75. So we set it at 49 = 24 + 25.
	
	gpio_set_mode(GPIO_CLKGEN, GPIO_MODE_ALT_FUNC, 0);

	SOC_CLKGEN[1].CTRL = SET_CLKGEN_LO(23) | SET_CLKGEN_HI(24); // 24.375 mhz

	SOC_CLK.ENABLE = CLK_CLKGEN1 | CLK_SSI;
#else
	SOC_CLK.ENABLE = CLK_SSI;
#endif

	// Setup the SSI interface for AIC23B "DSP-mode" format,
	// 16 bit stereo operation, 16 khz. 
	//
	// The TRIG value + 1 is the number of HCLK cycles between
	// each frame. A frame in DSP-mode is 2*16 bits (one stereo sample pair).
	// Thus TRIG + 1 = HCLK / 16000 => TRIG = 4687 (Argus-3).
	//
	// PERIOD controls the clock speed of the actual serial data.
	// It must be slower than 20 mhz according to the AIC23B datasheet
	// and it must be fast enough so that 33*tperiod > 1/fs.
	// (The number of data bits is 32, plus 1 cycle for the frame pulse)
	//
	// If we set PERIOD = 9, the cycle time will be 2*(9+1)/HCLK = 267 ns (Argus-3)
	// well above the 50 ns constraint, 
	// giving a total frame transfer time of 33 * 267 ns = 8.8 us.
	// 1/fs = 1 / 16000 = 62.5 us, so we're well below that constraint.

	SOC_SSI.SSIA = 
		SET_SSIA_TRIG(BASE_CLOCK / 16000 - 1) |
		SET_SSIA_PERIOD(9) |
		SET_SSIA_WIDTH(15); // mono operation read only half the data.
		
	// The internal CLK of the SSI block in Argus divides each frame
	// transfer into cycles with 2 phases in each, phase 0 and phase 1.
	// Phase 0 is while CLK is positive and phase 1 when it is negative.
	// Incoming data is sampled on the positive edge of CLK.
	//
	// AIC23B in slave-mode samples SSI_FRM/SSI_TX on the positive
	// edge of SSI_CLK, with a setup/hold requirement of 10 ns each,
	// and outputs to SSI_RX with the beginning of a negative slope
	// on SSI_CLK (it seems).
	//
	// We need to invert the SSI_CLK compared to the internal CLK
	// to satisfy the setup/hold requirement on SSI_FRM. However
	// to satisfy our own hold requirements on the returned data
	// on SSI_RX we need to revert the SSI_CLK during the dataphase
	// because the codec outputs with start on the negative edge
	// of SSI_CLK so if we had been inverted, we would sample
	// on that edge with possible hold violations since they
	// specify a propagation delay of 0 min and 20 max.
	//
	// We can let SSI_CLK idle while we idle to save some power.
	//
	// The DSP-mode needs a positive SSI_FRM for one cycle before
	// the data.

	SOC_SSI.SSIB =
		SET_SSIB_TXEN(0)  |  // Enable transmitter
		SET_SSIB_RXEN(0)  |  // Enable receiver
		SET_SSIB_IFRM(0)  |  // Idle FRM inactive
		SET_SSIB_ICLK0(0) |  // Idle CLK inactive
		SET_SSIB_ICLK1(0) |  // -=-
		SET_SSIB_FFRM(1)  |  // FRM active high during frame sync
		SET_SSIB_FCLK0(0) |  // 
		SET_SSIB_FCLK1(1) |  // -=-
		SET_SSIB_DFRM(0)  |  // FRM inactive during data transfer
		SET_SSIB_DCLK0(0) |  // 
		SET_SSIB_DCLK1(1) |  // -=-
		SET_SSIB_LBM(0)   |  // Disable loopback
		SET_SSIB_FINV(0);    // Don't toggle FRM each frame (I2S).

	// Finally activate the I/O pins (they might already be
	// activated by the kernelconfig at boot).
	//
	// Keep SSI_TX at zero until it is used, to avoid accidental built-in
	// digital loopback in the SSI-block.

	gpio_set_mode(GPIO_SSI_TX,  GPIO_MODE_OUT, 0);
	gpio_set_mode(GPIO_SSI_RX,  GPIO_MODE_ALT_FUNC, 0);
	gpio_set_mode(GPIO_SSI_FRM, GPIO_MODE_ALT_FUNC, 0);
	gpio_set_mode(GPIO_SSI_CLK, GPIO_MODE_ALT_FUNC, 0);
			
	printk("Argus SSI driver v2.1\n");

	// Disable the SSI clock until used to save power

#if USE_CLK_GATING
	SOC_CLK.DISABLE = CLK_SSI;
#endif
	return 0;
}

// Call init at boot or at module insertion.

module_init(ssi_init);
