/*
 * bootload.c
 * Simple boot loader for NAND chips on Etrax FS
 *
 * $Id: bootload.c,v 1.8 2006/11/28 10:05:39 ricardw Exp $
 * 
 */

#include <linux/types.h>
#include <linux/delay.h>

#include "mtd.h"
#include "nand.h"

#include <asm/arch/hwregs/reg_rdwr.h>
#include <asm/arch/hwregs/reg_map.h>
#include <asm/arch/hwregs/pinmux_defs.h>

#include "lib.h"

#define BOOT_ADDR  (CONFIG_ETRAX_PTABLE_SECTOR + 0x40200000)

/* bits for nand_rw() `cmd'; or together as needed */

#define NANDRW_READ		0x01
#define NANDRW_WRITE		0x00
#define NANDRW_JFFS2		0x02
#define NANDRW_JFFS2_SKIP	0x04

#define ROUND_DOWN(value, boundary)      ((value) & (~((boundary)-1)))

/* set $r8 to RAM_INIT_MAGIC, $r12 to NAND_BOOT_MAGIC then jump */
#define BOOT(addr)  __asm__ volatile (" \
	move.d 0x56902387, $r8\n\
	move.d 0x9A9DB001, $r12\n\
	jump %0\n\
	nop\n\
	" : : "r" (addr))

#define D(x) 

extern struct mtd_info *crisv32_nand_flash_probe(void);

extern int _end, _bss, _edata;

/* 
 * NAND read/write from U-Boot 1.4.4
 * Modified for newer mtd and use of mtd interface instead of nand directly.
 *
 * cmd: 0: NANDRW_WRITE			write, fail on bad block
 *	1: NANDRW_READ			read, fail on bad block
 *	2: NANDRW_WRITE | NANDRW_JFFS2	write, skip bad blocks
 *	3: NANDRW_READ | NANDRW_JFFS2	read, data all 0xff for bad blocks
 *      7: NANDRW_READ | NANDRW_JFFS2 | NANDRW_JFFS2_SKIP read, skip bad blocks
 */
static int
nand_rw (struct mtd_info* mtd, int cmd,
	 size_t start, size_t len,
	 size_t *retlen, u_char * buf)
{
	int ret = 0, n, total = 0;

	/* eblk (once set) is the start of the erase block containing the
	 * data being processed.
	 */
	size_t eblk = ~0;	/* force mismatch on first pass */
	size_t erasesize = mtd->erasesize;

	while (len) {
		if ((start & (-erasesize)) != eblk) {
			/* have crossed into new erase block, deal with
			 * it if it is marked bad.
			 */
			eblk = start & (-erasesize); /* start of block */
			D(
			  puts("New block ");
			  putx(eblk);
			  putnl();
			 )
			if (mtd->block_isbad(mtd, eblk)) {
				if (cmd == (NANDRW_READ | NANDRW_JFFS2)) {
					while (len > 0 &&
					       start - eblk < erasesize) {
						*(buf++) = 0xff;
						++start;
						++total;
						--len;
					}
					continue;
				} else if (cmd == (NANDRW_READ | NANDRW_JFFS2 | NANDRW_JFFS2_SKIP)) {
					start += erasesize;
					continue;
				} else if (cmd == (NANDRW_WRITE | NANDRW_JFFS2)) {
					/* skip bad block */
					start += erasesize;
					continue;
				} else {
					ret = 1;
					break;
				}
			}
		}
		/* The ECC will not be calculated correctly if
		   less than 512 is written or read */
		/* Is request at least 512 bytes AND it starts on a proper boundry */
		if((start != ROUND_DOWN(start, 0x200)) || (len < 0x200))
			puts("Warning block writes should be at least 512 bytes and start on a 512 byte boundry\r\n");

#if 0
		buf = (void *) 0x38008000; /* to fixed address, for testing */
#endif

		if (cmd & NANDRW_READ) {
			ret = mtd->read_ecc(mtd, start, 
					    min(len, eblk + erasesize - start),
					    (size_t *)&n, (u_char*)buf, 
					    NULL, NULL);
		} else {
			ret = mtd->write_ecc(mtd, start,
					     min(len, eblk + erasesize - start),
					     (size_t *)&n, (u_char*)buf, 
					     NULL, NULL);
		}

		if (ret) {
			break;
		}

		start  += n;
		buf   += n;
		total += n;
		len   -= n;
	}
	if (retlen)
		*retlen = total;

	return ret;
}


void
bootload()
{
	char revision;
	struct mtd_info *mtd;

	serial_init();

	__asm__ volatile ("move $vr,%0" : "=rm" (revision));
	if (revision < 32)
	{
		puts("You need an ETRAX FS to run Linux 2.6/crisv32.\r\n");
		while(1);
	}

	puts("\r\n\nETRAX FS NAND boot loader\r\n");
	puts("=========================\r\n");
	puts("Rev 1, " __DATE__ " " __TIME__ "\r\n");

	puts("CPU revision: ");
	putx(revision);
	putnl();

	puts("Bootloader main at ") ;
	putx((int) bootload);
	putnl();

	puts("Data end: ");
	putx((long) &_edata);
	putnl();

	puts("Bss: ");
	putx((long) &_bss);
	putnl();

	puts("Heap: ");
	putx((long) &_end);
	putnl();

#if 0 /* loop calibration */
	volatile int i;
	puts ("10000 loops...");
	for (i = 0; i < 10000; i++)
		udelay(1000);
	puts("done\r\n");
#endif

	puts("Identifying nand chip...\r\n");
	mtd = crisv32_nand_flash_probe();
	puts("Done.\r\n");
	
	if (mtd) {
		puts("Chip identified. ");
#if 0 /* print chip parameters */
		if (mtd->name)
			puts(mtd->name);

		puts("\r\ntype: ");
		putx(mtd->type);
		puts("\r\nflags: ");
		putx(mtd->flags);
		puts("\r\nsize: ");
		putx(mtd->size);
		puts("\r\nerasesize: ");
		putx(mtd->erasesize);
		puts("\r\noobblock: ");
		putx(mtd->oobblock);
		puts("\r\noobsize: ");
		putx(mtd->oobsize);
		puts("\r\necctype: ");
		putx(mtd->ecctype);
		puts("\r\neccsize: ");
		putx(mtd->eccsize);
#endif
		putnl();

		puts("Bad blocks:\r\n");

		int i;
		for (i = 0; i < mtd->size; i += mtd->erasesize) {
			if (mtd->block_isbad(mtd, i)) {
				putx(i);
				putnl();
			}
		}

#if 0 /* print oob parameters */
		puts("Oob info:\r\n");
		puts("useecc: ");
		putx(mtd->oobinfo.useecc);
		puts("\r\neccbytes: ");
		putx(mtd->oobinfo.eccbytes);
		puts("\r\neccpos: ");
		for (i = 0; i < mtd->oobinfo.eccbytes; i++) {
			putx(mtd->oobinfo.eccpos[i]);
			putc(' ');
		}
		putnl();
#endif

		puts("Bootload in progress...");
		int res, copied;
		res = nand_rw(mtd, 
			      NANDRW_READ | NANDRW_JFFS2 | NANDRW_JFFS2_SKIP, 
			      CONFIG_ETRAX_PTABLE_SECTOR, 
			      0x200000, /* 2 megs */
			      &copied, 
			      (void *) BOOT_ADDR);

		puts("complete, status ");
		putx(res);
		puts(", loaded ");
		putx(copied);
		puts(" bytes\r\n");
#if 1
		puts("Data in DRAM:\r\n");
		putx(* (int *) (BOOT_ADDR + 0));
		putc(' ');
		putx(* (int *) (BOOT_ADDR + 4));
		putc(' ');
		putx(* (int *) (BOOT_ADDR + 8));
		putnl();
#endif

		if (res) 
			error("Corrupt data in NAND flash.");
		else
		{
			puts("Booting...\r\n");
			BOOT(BOOT_ADDR);
		}
	} else
		error("No NAND flash chip found to boot from.");

	while (1)
		; /* hang around until hell freezes over */
}
