/*
 * Somewhat generic "board-side" code to support SPI drivers for chips
 * with a CRIS v32 and later.  Not really board-specific, and only for
 * registration of SPI devices for MMC, hence the "mmcspi" part of the
 * name instead of a proper board name.
 *
 * Copyright (c) 2007 Axis Communications AB
 *
 * TODO: SDIO interrupt-pin support (though can't be done until
 * there's support added in both the mmc_spi and the mmc frameworks).
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <linux/types.h>
#include <linux/platform_device.h>
#include <linux/spi/spi.h>
#include <linux/spi/mmc_spi.h>
#include <linux/mmc/host.h>
#include <asm/arch/board.h>
#include <asm/arch/pinmux.h>
#include <asm/arch/dma.h>
#include <linux/err.h>

#include <asm/io.h>

/* We need some housekeeping.  */
#define CONCAT_(a, b) a ## b
#define CONCAT(a, b) CONCAT_(a, b)
#define CONCAT3(a, b, c) CONCAT(CONCAT(a, b), c)

/* Grr.  Why not define them as usual in autoconf, #ifdef REVEAL_MODULES?  */
#if !defined(CONFIG_SPI_ETRAX_SSER) && defined(CONFIG_SPI_ETRAX_SSER_MODULE)
#define CONFIG_SPI_ETRAX_SSER
#endif

#if !defined(CONFIG_ETRAX_SPI_SSER0) && defined(CONFIG_ETRAX_SPI_SSER0_MODULE)
#define CONFIG_ETRAX_SPI_SSER0
#endif

#if !defined(CONFIG_ETRAX_SPI_SSER1) && defined(CONFIG_ETRAX_SPI_SSER1_MODULE)
#define CONFIG_ETRAX_SPI_SSER1
#endif

#if !defined(CONFIG_SPI_ETRAX_GPIO) && defined(CONFIG_SPI_ETRAX_GPIO_MODULE)
#define CONFIG_SPI_ETRAX_GPIO
#endif

#define CONFIGURED_PIN(x) ((x) != NULL && *(x) != 0 && *(x) != ' ')

/* Helper function to configure an iopin for input.  */

static int crisv32_config_pin_in(struct crisv32_iopin *pin, const char *name)
{
	int ret = crisv32_io_get_name(pin, name);
	if (ret)
		return ret;

	crisv32_io_set_dir(pin, crisv32_io_dir_in);
	return 0;
}

/*
 * Writable data pointed to by the constant parts of each MMC_SPI bus
 * interface.  Should only hold MMC-specific state (for the
 * MMC-specific pins), nothing SPI-generic.
 */

struct crisv32_mmc_spi_pinstate {
	struct crisv32_iopin write_protect_pin;
	struct crisv32_iopin card_detect_pin;

	/* We must poll for card-detect, which we do each:  */
#define CARD_DETECT_CHECK_INTERVAL (2*HZ)
	/* We poll using a self-arming workqueue function.  */
	struct work_struct cd_work;

	/* When we detect a change in card presence, we call this
	 * function, supplied by the mmc_spi framework: */
	irqreturn_t (*detectfunc)(int, void *);

	/*
	 * We call that function with this parameter as the second
	 * one.  We must assume the other parameters are unused, as we
	 * don't have an interrupt context.
	 */
	void *detectfunc_param2;

	/* State for the card-detect worker. */
	int card_present;
};

/* Rearming "work" function for card-detect polling.  */

static void crisv32_cd_worker(void *x)
{
	struct crisv32_mmc_spi_pinstate *state = x;

	/* It's a pull-up, so a card is present when 0.  */
	int card_present = crisv32_io_rd(&state->card_detect_pin) == 0;

	if (card_present != state->card_present) {
		state->card_present = card_present;
		state->detectfunc(0, state->detectfunc_param2);
	}
	schedule_delayed_work(&state->cd_work, CARD_DETECT_CHECK_INTERVAL);
}

/* The parts of MMC-specific configuration common to GPIO and SSER.  */

static int crisv32_mmcspi_config_common(struct spi_device *spidev,
					irqreturn_t (*intfunc)(int, void *),
				struct mmc_host *mmc_host,
				const struct crisv32_mmc_spi_pindata *pd)
{
	int ret = 0;
	struct crisv32_mmc_spi_pinstate *ps = pd->pinstate;

	/*
	 * If we don't have a card-detect pin, the card must be
	 * present at startup (including insmod/modprobe).  No
	 * re-scans are done if no card is present at that time.
	 */
	if (CONFIGURED_PIN(pd->card_detect)) {
		ret = crisv32_config_pin_in(&ps->card_detect_pin,
					    pd->card_detect);

		if (ret != 0)
			goto bad_card_detect;

		/* Need to cast away const from pd, unfortunately.  */
		INIT_WORK(&ps->cd_work, crisv32_cd_worker, (void *) ps);
		ps->card_present = 1;
		ps->detectfunc = intfunc;
		ps->detectfunc_param2 = mmc_host;
		crisv32_cd_worker(ps);
	} else
		ps->detectfunc = NULL;

	/*
	 * We set up for checking for presence of a write-protect pin
	 * as pin.port being non-NULL.
	 */
	memset(&ps->write_protect_pin, 0, sizeof ps->write_protect_pin);
	if (CONFIGURED_PIN(pd->write_protect)) {
		ret = crisv32_config_pin_in(&ps->write_protect_pin,
					    pd->write_protect);

		if (ret != 0)
			goto bad_write_protect;
	}

	return 0;

 bad_write_protect:
	if (ps->detectfunc) {
		cancel_rearming_delayed_work(&ps->cd_work);
		ps->detectfunc = NULL;
	}

 bad_card_detect:
	return ret;
}

/* A function to undo crisv32_mmcspi_config_common.  */

static void crisv32_mmcspi_deconfig_common(const struct
					   crisv32_mmc_spi_pindata *pd)
{
	if (pd->pinstate->detectfunc) {
		cancel_rearming_delayed_work(&pd->pinstate->cd_work);
		pd->pinstate->detectfunc = NULL;
	}

	/* We don't have to undo the pin allocations, being GPIO.  */
}

/* Helper function for write-protect sense.  */

static int crisv32_mmcspi_get_ro_helper(struct crisv32_mmc_spi_pinstate *ps)
{
	return ps->write_protect_pin.port != NULL
		&& crisv32_io_rd(&ps->write_protect_pin) != 0;
}

/* The hardware-interface-specific SPI+MMC parts.  */

#ifdef CONFIG_SPI_ETRAX_SSER
static const char crisv32_matching_spi_sser_driver_name[] __init_or_module
 = "spi_crisv32_sser";

/*
 * Make up something we can use in tables and function without
 * #ifdef-cluttering.
 */
#ifdef CONFIG_ETRAX_SPI_SSER0_DMA
#define WITH_ETRAX_SPI_SSER0_DMA 1
#else
#define WITH_ETRAX_SPI_SSER0_DMA 0
#endif
#ifdef CONFIG_ETRAX_SPI_SSER1_DMA
#define WITH_ETRAX_SPI_SSER1_DMA 1
#else
#define WITH_ETRAX_SPI_SSER1_DMA 0
#endif

/* Write-protect sense for the SSER interface.  */

static int crisv32_mmcspi_sser_get_ro(struct device *dev)
{
	struct spi_device *spidev = to_spi_device(dev);
	struct crisv32_mmc_spi_sser_hwdata *sser_defs
		= spidev->controller_data;
	struct crisv32_mmc_spi_pindata *pd = &sser_defs->mmc;
	return crisv32_mmcspi_get_ro_helper(pd->pinstate);
}

/* Initialize the MMC-specific parts of the MMC+SPI interface, SSER.  */

static int crisv32_mmcspi_sser_init(struct device *dev,
				    irqreturn_t (*intfunc)(int, void *),
				    void *xmmc_host)
{
	struct mmc_host *mmc_host = xmmc_host;
	struct spi_device *spidev = to_spi_device(dev);
	struct crisv32_mmc_spi_sser_hwdata *sser_defs
		= spidev->controller_data;
	int ret = crisv32_mmcspi_config_common(spidev, intfunc, mmc_host,
					       &sser_defs->mmc);
	if (ret != 0)
		return ret;

	mmc_host->f_max = spidev->max_speed_hz;

	/*
	 * Let's just set this to ceil(100e6/65536).  It wouldn't be
	 * hard to change the base frequency to support down to 450Hz,
	 * but there's no apparent requirement from known hardware.
	 * Would also require adjustments to the SPI code, not just
	 * here.
	 *
	 * On second thought, let's not set f_min unconditionally, as
	 * mmc doesn't treat it as a limit, but as the frequency to
	 * use at initialization.  We stick with what mmc_spi sets, if
	 * it fits.
	 */
	if (mmc_host->f_min < 1526)
		mmc_host->f_min = 1526;

	dev_info(dev, "CRIS v32 mmc_spi support hooked to SPI on sser%d"
		 " (cd: %s, wp: %s)\n",
		 spidev->master->bus_num,
		 CONFIGURED_PIN(sser_defs->mmc.card_detect)
		 ? sser_defs->mmc.card_detect : "(none)",
		 CONFIGURED_PIN(sser_defs->mmc.write_protect)
		 ? sser_defs->mmc.write_protect : "(none)");
	return 0;
}

/* Similarly, to undo crisv32_mmcspi_sser_init.  */

static void crisv32_mmcspi_sser_exit(struct device *dev, void *xmmc_host)
{
	struct spi_device *spidev = to_spi_device(dev);
	struct crisv32_mmc_spi_sser_hwdata *sser_defs
		= spidev->controller_data;

	crisv32_mmcspi_deconfig_common(&sser_defs->mmc);
}

/*
 * Connect sser and DMA channels and return the proper numbers to use.
 *
 * This function exists really only to avoid having the following constants
 * tabled: regi_sserN, SSERn_INTR_VECT, pinmux_sserN, SYNC_SERn_TX_DMA_NBR,
 * SYNC_SERn_RX_DMA_NBR dma_sserN, regi_dmaM, regi_dmaN, DMAm_INTR_VECT,
 * DMAn_INTR_VECT.  I might have missed some. :-)
 */
static int crisv32_allocate_sser(struct crisv32_regi_n_int *sserp,
				 struct crisv32_regi_n_int *dmainp,
				 struct crisv32_regi_n_int *dmaoutp,
				 u32 sserno)
{
	int ret;
	u32 regi_sser = sserno == 0 ? regi_sser0 : regi_sser1;
	u32 sser_irq = sserno == 0 ? SSER0_INTR_VECT : SSER1_INTR_VECT;
	BUG_ON(sserno > 1 || sserp == NULL);

	ret = crisv32_pinmux_alloc_fixed(sserno == 0
					 ? pinmux_sser0 : pinmux_sser1);
	if (ret != 0)
		return ret;

	sserp->regi = regi_sser;
	sserp->irq = sser_irq;

	if (dmaoutp) {
		crisv32_request_dma(sserno == 0
				    ? SYNC_SER0_TX_DMA_NBR
				    : SYNC_SER1_TX_DMA_NBR,
				    "SD/MMC SPI dma tr",
				    DMA_VERBOSE_ON_ERROR | DMA_PANIC_ON_ERROR,
				    /* Let's be brave and ask for the
				       max.  Except it's simplex, so
				       let's request half the max
				       speed in either direction.  */
				    50 * 1000 * 1000 / 8 / 2,
				    sserno == 0 ? dma_sser0 : dma_sser1);
		dmaoutp->regi = sserno == 0
			? CONCAT(regi_dma, SYNC_SER0_TX_DMA_NBR)
			: CONCAT(regi_dma, SYNC_SER1_TX_DMA_NBR);
		dmaoutp->irq = sserno == 0
			? CONCAT3(DMA, SYNC_SER0_TX_DMA_NBR, _INTR_VECT)
			: CONCAT3(DMA, SYNC_SER1_TX_DMA_NBR, _INTR_VECT);
	}

	if (dmainp) {
		crisv32_request_dma(sserno == 0
				    ? SYNC_SER0_RX_DMA_NBR
				    : SYNC_SER1_RX_DMA_NBR,
				    "SD/MMC SPI dma rec",
				    DMA_VERBOSE_ON_ERROR | DMA_PANIC_ON_ERROR,
				    50 * 1000 * 1000 / 8 / 2,
				    sserno == 0 ? dma_sser0 : dma_sser1);
		dmainp->regi = sserno == 0
			? CONCAT(regi_dma, SYNC_SER0_RX_DMA_NBR)
			: CONCAT(regi_dma, SYNC_SER1_RX_DMA_NBR);
		dmainp->irq = sserno == 0
			? CONCAT3(DMA, SYNC_SER0_RX_DMA_NBR, _INTR_VECT)
			: CONCAT3(DMA, SYNC_SER1_RX_DMA_NBR, _INTR_VECT);
	}

	return 0;
}

/* Undo whatever allocations et. al. that crisv32_allocate_sser did.  */

static void crisv32_free_sser(u32 sserno, int with_dma)
{
	int ret;

	if (with_dma) {
		crisv32_free_dma(sserno == 0
				 ? SYNC_SER0_RX_DMA_NBR
				 : SYNC_SER1_RX_DMA_NBR);
		crisv32_free_dma(sserno == 0
				 ? SYNC_SER0_TX_DMA_NBR
				 : SYNC_SER1_TX_DMA_NBR);
	}

	ret = crisv32_pinmux_dealloc_fixed(sserno == 0
					   ? pinmux_sser0 : pinmux_sser1);

	if (ret != 0)
		panic("%s: crisv32_pinmux_dealloc_fixed returned %d\n",
		       __FUNCTION__, ret);
}

/* Convenience-macro to avoid typos causing diffs between sser ports.  */

#define SSER_CDATA(n)							\
	{								\
		{							\
			.using_dma = WITH_ETRAX_SPI_SSER##n##_DMA,	\
			.iface_allocate = crisv32_allocate_sser##n,	\
			.iface_free = crisv32_free_sser##n		\
		},							\
		{							\
			.card_detect					\
				= CONFIG_ETRAX_SPI_MMC_CD_SSER##n##_PIN,\
			.write_protect					\
				= CONFIG_ETRAX_SPI_MMC_WP_SSER##n##_PIN,\
			.pinstate					\
				= &crisv32_mmcspi_sser##n##_pinstate	\
		}							\
	}

#ifdef CONFIG_ETRAX_SPI_SSER0

/* Allocate hardware to go with sser0 and fill in the right numbers.  */

static int crisv32_allocate_sser0(struct crisv32_regi_n_int *sserp,
				  struct crisv32_regi_n_int *dmainp,
				  struct crisv32_regi_n_int *dmaoutp)
{
	return crisv32_allocate_sser(sserp, dmainp, dmaoutp, 0);
}

/* Undo those allocations.  */

static void crisv32_free_sser0(void)
{
	crisv32_free_sser(0, WITH_ETRAX_SPI_SSER0_DMA);
}

static struct crisv32_mmc_spi_pinstate crisv32_mmcspi_sser0_pinstate;
static const struct crisv32_mmc_spi_sser_hwdata crisv32_mmcspi_sser0_cdata
 = SSER_CDATA(0);
#endif /* CONFIG_ETRAX_SPI_SSER0 */

#ifdef CONFIG_ETRAX_SPI_SSER1

/* Allocate hardware to go with sser0 and fill in the right numbers.  */

static int crisv32_allocate_sser1(struct crisv32_regi_n_int *sserp,
				  struct crisv32_regi_n_int *dmainp,
				  struct crisv32_regi_n_int *dmaoutp)
{
	return crisv32_allocate_sser(sserp, dmainp, dmaoutp, 1);
}

/* Undo those allocations.  */

static void crisv32_free_sser1(void)
{
	crisv32_free_sser(1, WITH_ETRAX_SPI_SSER1_DMA);
}

static struct crisv32_mmc_spi_pinstate crisv32_mmcspi_sser1_pinstate;
static const struct crisv32_mmc_spi_sser_hwdata crisv32_mmcspi_sser1_cdata
 = SSER_CDATA(1);

#endif /* CONFIG_ETRAX_SPI_SSER1 */

static struct mmc_spi_platform_data crisv32_mmcspi_sser_pdata = {
	.init = crisv32_mmcspi_sser_init,
	.exit = __devexit_p(crisv32_mmcspi_sser_exit),
	.detect_delay = 0,
	.get_ro = crisv32_mmcspi_sser_get_ro,
	.ocr_mask = MMC_VDD_32_33|MMC_VDD_33_34,
	.setpower = NULL
};

#endif /* CONFIG_SPI_ETRAX_SSER */

#ifdef CONFIG_SPI_ETRAX_GPIO

static const char crisv32_matching_spi_gpio_driver_name[] __init_or_module
 = "spi_crisv32_gpio";

/* Write-protect sense for the GPIO interface.  */

static int crisv32_mmcspi_gpio_get_ro(struct device *dev)
{
	struct spi_device *spidev = to_spi_device(dev);
	struct crisv32_mmc_spi_gpio_hwdata *gpio_defs
		= spidev->controller_data;
	struct crisv32_mmc_spi_pindata *pd = &gpio_defs->mmc;

	return crisv32_mmcspi_get_ro_helper(pd->pinstate);
}

/* Initialize the MMC-specific parts of the MMC+SPI interface, GPIO.  */

static int crisv32_mmcspi_gpio_init(struct device *dev,
				    irqreturn_t (*intfunc)(int, void *),
				    void *xmmc_host)
{
	struct mmc_host *mmc_host = xmmc_host;
	struct spi_device *spidev = to_spi_device(dev);
	struct crisv32_mmc_spi_gpio_hwdata *gpio_defs
		= spidev->controller_data;
	int ret = crisv32_mmcspi_config_common(spidev, intfunc, mmc_host,
					       &gpio_defs->mmc);
	if (ret)
		return ret;

	mmc_host->f_max = spidev->max_speed_hz;

	/*
	 * We don't set f_min, as the mmc code doesn't treat it as a
	 * limit, but as the frequency to use at initialization.  We
	 * stick with what mmc_spi sets.
	 */
	dev_info(dev, "CRIS v32 mmc_spi support hooked to SPI on GPIO"
		 " (bus #%d, cd: %s, wp: %s)\n",
		 spidev->master->bus_num,
		 CONFIGURED_PIN(gpio_defs->mmc.card_detect)
		 ? gpio_defs->mmc.card_detect : "(none)",
		 CONFIGURED_PIN(gpio_defs->mmc.write_protect)
		 ? gpio_defs->mmc.write_protect : "(none)");
	return 0;
}

/* Similarly, to undo crisv32_mmcspi_gpio_init.  */

static void crisv32_mmcspi_gpio_exit(struct device *dev, void *xmmc_host)
{
	struct spi_device *spidev = to_spi_device(dev);
	struct crisv32_mmc_spi_gpio_hwdata *gpio_defs
		= spidev->controller_data;

	crisv32_mmcspi_deconfig_common(&gpio_defs->mmc);
}

static const struct mmc_spi_platform_data crisv32_mmcspi_gpio_pdata = {
	.init = crisv32_mmcspi_gpio_init,
	.exit = __devexit_p(crisv32_mmcspi_gpio_exit),
	.detect_delay = 0,
	.get_ro = crisv32_mmcspi_gpio_get_ro,
	.ocr_mask = MMC_VDD_32_33|MMC_VDD_33_34,
	.setpower = NULL
};

static struct crisv32_mmc_spi_pinstate crisv32_mmcspi_gpio_pinstate;
static const struct crisv32_mmc_spi_gpio_hwdata crisv32_mmcspi_gpio_cdata = {
	{
		.cs	= CONFIG_ETRAX_SPI_CS_PIN,
		.miso	= CONFIG_ETRAX_SPI_DATAIN_PIN,
		.mosi	= CONFIG_ETRAX_SPI_DATAOUT_PIN,
		.sclk	= CONFIG_ETRAX_SPI_CLK_PIN
	},
	{
		.card_detect	= CONFIG_ETRAX_SPI_MMC_CD_GPIO_PIN,
		.write_protect	= CONFIG_ETRAX_SPI_MMC_WP_GPIO_PIN,
		.pinstate	= &crisv32_mmcspi_gpio_pinstate
	}
};
#endif /* CONFIG_SPI_ETRAX_GPIO */

struct crisv32_s_b_i_and_driver_info {
	struct spi_board_info sbi;
	const char *driver_name;

	/*
	 * We have to adjust the platform device at time of creation
	 * below, to allow the Linux DMA framework to handle DMA.  In
	 * the name of simplicity, we state the dma:able property
	 * twice, a bit redundantly; here and in the controller_data
	 * structure pointed to by spi_board_info.
	 */
	int dma_able;
};

/* Convenience-macro to avoid typos causing diffs between sser ports.  */

#define SSER_CONFIG(n)							\
	{								\
		/*							\
		 * No meaningful values for .irq or .chip_select,	\
		 * so we leave them out.				\
		 */							\
		{							\
			.modalias		= "mmc_spi",		\
			.platform_data					\
				 = &crisv32_mmcspi_sser_pdata,		\
			/*						\
			 * Casting to avoid a compiler const-warning.	\
			 */						\
			.controller_data				\
				= ((void *)				\
				   &crisv32_mmcspi_sser##n##_cdata),	\
			.max_speed_hz		= 50 * 1000 * 1000,	\
			.bus_num		= n,			\
			/*						\
			 * We only provide one mode, so it must be	\
			 * default.					\
			 */						\
			.mode			= SPI_MODE_3		\
		},							\
		crisv32_matching_spi_sser_driver_name,			\
		WITH_ETRAX_SPI_SSER##n##_DMA				\
	}

static const struct crisv32_s_b_i_and_driver_info
crisv32_mmcspi_devices[] __initdata = {
#ifdef CONFIG_ETRAX_SPI_SSER0
	SSER_CONFIG(0),
#endif
#ifdef CONFIG_ETRAX_SPI_SSER1
	SSER_CONFIG(1),
#endif
#ifdef CONFIG_SPI_ETRAX_GPIO
	{
		{
			.modalias		= "mmc_spi",
			.platform_data		= &crisv32_mmcspi_gpio_pdata,
			/* Casting to avoid a compiler const-warning.  */
			.controller_data
				= (void *) &crisv32_mmcspi_gpio_cdata,

			/* No, we can't even reach this number with GPIO.  */
			.max_speed_hz		= 5 * 1000 * 1000,
			.bus_num		= 2
		},
		crisv32_matching_spi_gpio_driver_name,
		0
	},
#endif
};

/* Must not be __initdata.  */
static const char controller_data_ptr_name[] = "controller_data_ptr";
static u64 allmem_mask = ~(u32) 0;

/*
 * We have to register the SSER and GPIO "platform_device"s separately
 * from the "spi_board_info"s in order to have the drivers separated
 * from the hardware instance info.
 */

static int __init crisv32_register_mmcspi(void)
{
	int ret;
	void *retp;
	const struct crisv32_s_b_i_and_driver_info *sbip;

	for (sbip = crisv32_mmcspi_devices;
	     sbip < crisv32_mmcspi_devices
		     + ARRAY_SIZE(crisv32_mmcspi_devices);
	     sbip++) {

		/*
		 * We have to pass the controller data as a device
		 * "resource" (FIXME: too?), so it can be available
		 * for the setup *before* starting the SPI core -
		 * which is where .controller_data makes it to the SPI
		 * structure!
		 */
		struct resource mmcspi_res = {
			.start = (resource_size_t) sbip->sbi.controller_data,
			.end = (resource_size_t) sbip->sbi.controller_data,
			.name = controller_data_ptr_name
		};

		/*
		 * Presumably we could pass the irqs and register
		 * addresses and... as individual resources here
		 * instead of privately in spi_board_info
		 * .controller_data, and make that part a bit more
		 * readable.  Maybe not worthwhile.
		 */
		/* Need to cast away const here.  FIXME: fix the pdrs API.  */
		retp = platform_device_alloc(sbip->driver_name,
					     sbip->sbi.bus_num);
		if (IS_ERR_VALUE(PTR_ERR(retp)))
			return PTR_ERR(retp);

		/*
		 * We can't use platform_device_register_simple as we
		 * want to set stuff in the device to mark that we'd
		 * prefer buffers explicitly passed as DMA-able.
		 * Disabling/omitting the "if" below, cause testing of
		 * the non-DMA-mem execution path of a DMA-capable
		 * driver.
		 */
		if (sbip->dma_able) {
			struct platform_device *pdev = retp;

			pdev->dev.dma_mask = &allmem_mask;
			pdev->dev.coherent_dma_mask = allmem_mask;
		}

		if ((ret = platform_device_add_resources(retp, &mmcspi_res, 1)) != 0
		    || (ret = platform_device_add(retp)) != 0) {
			platform_device_put(retp);
			return ret;
		}

		/*
		 * The cost of keeping all info rooted at
		 * crisv32_mmcspi_devices is the allocation overhead
		 * for allocating separately each board info (in the
		 * unlikely event that there's more than one).
		 */
		ret = spi_register_board_info(&sbip->sbi, 1);
		if (ret != 0)
			return ret;
	}

	return 0;
}
#ifdef MODULE
#error "Non-module because spi_register_board_info is one-way; there's no unregister function"
#endif

/*
 * Needs to be called before __initcall/module_init because the SPI
 * bus scanning happens when the actual driver registers with the SPI
 * machinery (spi_bitbang_start/spi_register_master), not when the
 * device registers (spi_register_board_info).
 */
subsys_initcall(crisv32_register_mmcspi);
