##=============================================================================
##
##      nand_init.S
##
##      The bootrom copies data from the NAND flash to the internal RAM but
##      due to a bug/feature we can only trust the 256 first bytes. So this
##      code copies more data from NAND flash to internal RAM. Obvioulsy this
##      code must fit in the first 256 bytes so alter with care.
##
##	Some notes about the bug/feature for future reference:
##        The bootrom copies the first 127 KB from NAND flash to internal
##        memory. The problem is that it does a bytewise copy, copying
##        a single byte from the lowest byte of the bus for each address.
##        NAND flashes autoincrement on the address so for a 16 bit device
##        each read/write increases the address by two. So the copy loop in the
##        bootrom will discard every second byte. This is solved by inserting
##        zeroes in every second byte in the first erase block, in order
##        to get contiguous code.
##
##        The bootrom also incorrectly assumes that it can read the flash
##        linearly with only one read command but the flash will actually
##        switch between normal area and spare area if you do that so we
##        can't trust more than the first 256 bytes.
##
##=============================================================================

#include <asm/arch/hwregs/asm/reg_map_asm.h>
#include <asm/arch/hwregs/asm/gio_defs_asm.h>
#include <asm/arch/hwregs/asm/pinmux_defs_asm.h>
#include <asm/arch/hwregs/asm/bif_core_defs_asm.h>
#include <asm/arch/hwregs/asm/config_defs_asm.h>

;; There are 8-bit NAND flashes and 16-bit NAND flashes.
;; We need to treat them slightly differently.
#if CONFIG_ETRAX_NANDFLASH_BUSWIDTH==2
#define PAGE_SIZE_ADDRESSES 256
#else
#define PAGE_SIZE_ADDRESSES 512
#endif

;; Block size for erase
#define ERASE_BLOCK 16384
#define PAGE_SIZE_BYTES 512

;; GPIO pins connected to NAND flash
#define CE 4
#define CLE 5
#define ALE 6
#define BY 7

;; Address space for NAND flash
#define NAND_RD_ADDR 0x90000000
#define NAND_WR_ADDR 0x94000000

#define READ_CMD 0x00
#define RESET_CMD 0xFF

;; Readability macros
#define CSP_MASK \
	REG_MASK(bif_core, rw_grp3_cfg, gated_csp0) | \
	REG_MASK(bif_core, rw_grp3_cfg, gated_csp1)
#define CSP_VAL \
	REG_STATE(bif_core, rw_grp3_cfg, gated_csp0, rd) | \
	REG_STATE(bif_core, rw_grp3_cfg, gated_csp1, wr)

;; Normally we initialize GPIO and bus interfaces.
;; This is strictly not necessary; boot ROM does this for us.
#define INTERFACE_SETUP (1)

;;----------------------------------------------------------------------------
;; Macros to set/clear GPIO bits

.macro SET x
	or.b   (1<<\x),$r9
	move.d $r9, [$r2]
.endm

.macro CLR x
	and.b  ~(1<<\x),$r9
	move.d $r9, [$r2]
.endm

.macro GPIO_SYNC
;; 	Originally, we read back data written to nand flash in order
;; 	to flush the pipeline. It turned out however, that the real
;;	culprit was a lack of wait states.
;;	This macro remains in the code however in case this conclusion
;;	is wrong too.
;;
;;	move.d [$r2], $r9 ; read back to flush pipeline
.endm

;;----------------------------------------------------------------------------
;; Read value from bus to temporary register to sync with previous write
;; This generates no signal to the NAND flash, since only chip select lines are
;; pulled out to the chip, and read is not gated with chip select for the write
;; area.

.macro BUS_SYNC r
	move.b [$r1], \r
.endm

;;----------------------------------------------------------------------------
;; Delay macro
;; x = delay = 10*x + 20 ns, e.g. DELAY 25 => 270ns delay, max 63 (650 ns)
;;(@200Mc)
;; r is temp reg used
;; Macro currently not used, save for a rainy day.

.macro DELAY x, r
	clear.d	\r
	addq	(\x),\r		; addq zero-extends its argument
7:	bne	7b
	subq	1, \r
.endm

;;----------------------------------------------------------------------------

copy_nand_to_ram:
	;; copy_nand_to_ram
	;; Arguments
	;;   r10 - destination
	;;   r11 - source offset
	;;   r12 - size
	;; Note : r10-r12 are clobbered on return
	;; Registers used:
	;;   r0 - NAND_RD_ADDR
	;;   r1 - NAND_WR_ADDR
	;;   r2 - reg_gio_rw_pa_dout
	;;   r3 - reg_gio_r_pa_din
	;;   r4 - tmp
	;;   r5 - byte counter within a page / tmp2
	;;   r6 - r_bootsel masked w/ 0x18
	;;   r7 - n/u
	;;   r8 - n/u
	;;   r9 - reg_gio_rw_pa_dout shadow
	move.d NAND_RD_ADDR, $r0
	move.d NAND_WR_ADDR, $r1
	move.d REG_ADDR(gio, regi_gio, rw_pa_dout), $r2
	move.d REG_ADDR(gio, regi_gio, r_pa_din), $r3

#if CONFIG_ETRAX_NANDFLASH_BUSWIDTH==2
	lsrq	1, $r11
#endif

#if INTERFACE_SETUP
 	;; Set up pinmux
	move.d REG_ADDR(pinmux, regi_pinmux, rw_pa), $r5
	move.d [$r5], $r4
	or.b 0xf0, $r4		; bits 4,5,6,7
	move.d $r4, [$r5]

	;; Set up GPIO
	move.d REG_ADDR(gio, regi_gio, rw_pa_oe), $r5
	move.d [$r5], $r4
	or.b (1<<ALE) | (1 << CLE) | (1<<CE), $r4
	move.d $r4, [$r5]

#endif
	;; Set up bif
	move.d REG_ADDR(bif_core, regi_bif_core, rw_grp3_cfg), $r5
	move.d CONFIG_ETRAX_MEM_GRP3_CONFIG, $r4 ; wait states
	and.d ~CSP_MASK, $r4
	or.d CSP_VAL, $r4
	move.d $r4, [$r5]

	move.d [$r2], $r9	; fetch PA DOUT to shadow register

	;; figure out how many address cycles the flash needs
	move.d REG_ADDR(config, regi_config, r_bootsel), $r5
	move.d [$r5], $r6
	andq 0x18, $r6 		; mask out bs3,4 (00=>3, 08=>4, 18=>5 cycles)

1:	;; Copy one page
	CLR CE
	SET CLE
	GPIO_SYNC
	moveq	READ_CMD, $r4
	move.b	$r4, [$r1]
	BUS_SYNC $r4
	CLR CLE
	SET ALE
	GPIO_SYNC
	clear.b [$r1] 		; Column address = 0
	move.d	$r11, $r4	; Address
	lsrq	9, $r4		; Row address is A9 and up
	move.b	$r4, [$r1]	; Row address byte #0
	lsrq	8, $r4
	cmpq	0x08, $r6	; 8 => Z, 0 => C, 18h => NZ,NC
	bcs	4f		; C (3 cycles) => jump
	move.b	$r4, [$r1]	; Row address byte #1 (DELAY SLOT) (no flagset)
	beq	3f		; Z (4 cycles) => jump
	lsrq	8, $r4		; (DELAY SLOT)
	move.b	$r4, [$r1]	; Row address byte #2 (5 cyc only)
	lsrq	8, $r4
3:
	move.b	$r4, [$r1]	; Row address byte #3 (5 cyc) or #2 (4 cyc)
4:
	BUS_SYNC $r4
	CLR ALE
	GPIO_SYNC

2:	move.d	[$r3], $r4
	and.d	1 << BY, $r4
	beq 2b
	nop
	movu.w  PAGE_SIZE_ADDRESSES, $r5
2:	; Copy one byte/word
#if CONFIG_ETRAX_NANDFLASH_BUSWIDTH==2
	move.w  [$r0], $r4
#else
	move.b  [$r0], $r4
#endif
	subq	1, $r5
	bne	2b
#if CONFIG_ETRAX_NANDFLASH_BUSWIDTH==2
	move.w  $r4, [$r10+]
#else
	move.b  $r4, [$r10+]
#endif
	subu.w	PAGE_SIZE_BYTES, $r12
	bhi	1b
	addu.w	PAGE_SIZE_ADDRESSES, $r11

	SET CE

	;; End of copy
