/*
 *  linux/drivers/mmc/sdio.c
 *
 *  Copyright (C) 2006 Pierre Ossman, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/mmc/card.h>
#include <linux/mmc/protocol.h>
#include <linux/mmc/sdio.h>
#include <linux/mmc/ids.h>

#define SDIO_ADDR_MASK 0x1FFFF

static unsigned char __sdio_readb(const struct mmc_card *card,
				  unsigned int func, unsigned int addr)
{
	struct mmc_command cmd;
	int err;

	BUG_ON(addr & ~SDIO_ADDR_MASK);
	BUG_ON(func > 7);

	cmd.opcode = SD_IO_RW_DIRECT;
	cmd.arg = (func << 28) | (addr << 9);
	cmd.flags = MMC_RSP_R5 | MMC_CMD_AC;

	err = mmc_wait_for_cmd(card->host, &cmd, 0);
	if (err != MMC_ERR_NONE)
		return 0xFF;

	if (cmd.resp[0] & (R5_ERROR | R5_FUNCTION_NUMBER | R5_OUT_OF_RANGE))
		return 0xFF;

	return cmd.resp[0] & 0xFF;
}

static void __sdio_writeb(const struct mmc_card *card, unsigned int func,
			  unsigned int addr, unsigned char data)
{
	struct mmc_command cmd;

	BUG_ON(addr & ~SDIO_ADDR_MASK);
	BUG_ON(func > 7);

	cmd.opcode = SD_IO_RW_DIRECT;
	cmd.arg = (1<<31) | (func << 28) | (addr << 9) | data;
	cmd.flags = MMC_RSP_R5 | MMC_CMD_AC;

	mmc_wait_for_cmd(card->host, &cmd, 0);
}

unsigned char sdio_readb(const struct mmc_function *func, unsigned int addr)
{
	return __sdio_readb(func->card, func->number, addr);
}

void sdio_writeb(const struct mmc_function *func,
		  unsigned int addr, unsigned char data)
{
	return __sdio_writeb(func->card, func->number, addr, data);
}

//EXPORT_SYMBOL_GPL(sdio_readb);
//EXPORT_SYMBOL_GPL(sdio_writeb);
EXPORT_SYMBOL(sdio_readb);
EXPORT_SYMBOL(sdio_writeb);

void sdio_discover_functions(struct mmc_card *card)
{
	int i;
	struct mmc_function *func;
	unsigned int cur_func, last_func;

	func = card->functions;
	cur_func = 1;
	last_func = card->n_functions;

	/* Memory card function isn't a real SDIO function */
	if (func->vendor == MMC_VENDOR_STORAGE) {
		func;
		last_func--; 
	}

	printk(KERN_INFO "SDIO: First byte: 0x%02x\n",
	       (int)__sdio_readb(card, 1, 0x3c));
}
